"""This module contains wrapper functions for plotting and visualizing data.

    @authors: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
              Jack Berry, Columbia University, jeb2242@columbia.edu
"""
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import seaborn as sns
from scipy.ndimage.filters import gaussian_filter
import scipy.sparse as sparse
import math
import analysis.analysis_utils as au

from itertools import combinations_with_replacement
from matplotlib.patches import Rectangle

#### Plotting functions for spatial activity ####

def plot_spatial_maps(mouse, neuron_data_df, m, neuron_list=None, norm_by_mouse=True, savefig=False,
                      filepath=None, filename=None, heatmap_bin_size=1, h=2, behavior=None, fmt='png', **kwargs):

    """
    Plots an event plot and ratemap for a subset of neurons in a given session

    TODO: finish commenting

    """
    
    #figsize = kwargs.get('figsize',(8,8))
    #cmap = kwargs.get('cmap','jet')
    #vmin = kwargs.get('vmin',0)
    neural_df = mouse.spikes
    beh_df = mouse.behavior[['X_bin', 'Y_bin']]
    df = pd.concat([neural_df, beh_df],axis=1)
    occupancy = df.groupby(['X_bin', 'Y_bin']).size()
    activity = df.groupby(['X_bin', 'Y_bin']).sum()
    x_max, y_max = behavior_boundaries(mouse.behavior.X_center, mouse.behavior.Y_center)

    max_rate = None

    if norm_by_mouse:
        max_rate = mouse_max_rate(activity_df=activity, occupancy_df=occupancy,
                                  heatmap_bin_size=heatmap_bin_size, h=h, fr=mouse.FR)

    if not neuron_list:
        neuron_list = mouse.spikes.columns

    for n in neuron_list:

        neural = mouse.spikes[n]
        behavior_df = mouse.behavior
        neuron_data = neuron_data_df.set_index(['mouse', 'neuron']).xs([m, n])

        plot_spatial(neural=neural, behavior_df=behavior_df, neuron_data=neuron_data, m=m, n=n,
                     heatmap_bin_size=heatmap_bin_size, savefig=savefig, filepath=filepath, filename=filename,
                     max_rate=max_rate, h=h, x_max=x_max, y_max=y_max, behavior=behavior, fmt=fmt,**kwargs)

def mouse_max_rate(activity_df, occupancy_df, heatmap_bin_size=1, h=2, fr=10):
    """
    Computes the maximum rate in any heatmap bin for the entire FOV. Intended to be used as a
    helper function for plot_spatial_maps. This is used to normalize the ratemap per mouse, as
    opposed to each neuron.

    TODO: finish commenting

    """
    activity_norm = activity_df.apply(lambda x: fr*x/occupancy_df)
    mouse_max = activity_norm.max().max()

    s = h/heatmap_bin_size
    mouse_gauss_max = (mouse_max/(s**2*math.pi))

    return mouse_gauss_max

def behavior_boundaries(x, y):
    """
    Defines the width and length of the behavioral arena, as defined by the mouse position. Intended to be used as a helper funtion
    for plot_spatial_maps

    TODO: finish commenting

    """

    x_max = x.max() - x.min()
    y_max = y.max() - y.min()

    return x_max, y_max


def plot_spatial(neural = None, behavior_df = None, neuron_data = None, m = None, n = None,
                 heatmap_bin_size = 1, h = 2, max_rate = None, savefig = False, filepath = None, 
                 filename = None, x_max = 50, y_max = 50, behavior = None, fmt='png',**kwargs):
    """
    Plots event map and rate map for a given neuron.

    TODO: finish commenting and make more general by taking out arm score

    Inputs:
        neural: Pandas Series
            The spike train for one neuron. 
        behavior_df: Pandas DataFrame
            Behavior DataFrame
        neuron_data: Pandas Series
            Relevant data
        m: str
            mouse name
        n: str
            neuron name
        filepath: str
            location to save figures
        heatmap_bin_size: int
            size (in cm) of bin size for plotting
        h: int
            smoothing kernel size (cm), genenrally the same as the bin size (for calculating spatial info)
        x_max: float
            width of arena, used for setting heatmap bin size
        y_max: float
            length of arena, usd for setting heatmap bin size
    
    """   

    df = pd.concat([neural, behavior_df],axis=1)
    occupancy = df.groupby(['X_bin','Y_bin']).size()
    activity = df.groupby(['X_bin','Y_bin']).sum()

    activity_norm = activity.apply(lambda x: 10*x/occupancy)

    x_bin = activity.reset_index().X_bin
    y_bin = activity.reset_index().Y_bin

    b = (x_max/heatmap_bin_size, y_max/heatmap_bin_size) 

    s = h/heatmap_bin_size #sigma =smoothing factor

    t= "{}, neuron {}.{}".format(m,n,fmt)

    #nipy_spectral

    if not filename:
        filename = "{}, EPM, neuron {}".format(m,n)

    filename1 = filename + ', eventmap.png'
    filename2 = filename + ', ratemap.png'

    #_=neuron_activity_plot(x, y, neural, t, savefig = savefig, filepath = filepath, filename = filename1)

    figsize = kwargs.get('figsize',(8,8))

    neuron_event_plot(df, neuron = n, behavior = behavior, title = t, savefig = savefig,
                      filepath = filepath, filename = filename1,**kwargs)
    
    cmap = kwargs.get('cmap','jet')
    vmin=kwargs.get('vmin',0)
    plot_heatmap(x_bin, y_bin, bins = b, weights = activity_norm[n], sigma = s, vmin = vmin, vmax = max_rate, 
                     cmap = cmap, title = t, savefig = savefig, filename = filename2, filepath = filepath,
                     figsize=figsize)


def neuron_event_plot(df, neuron, behavior = None, ax=None, **kwargs):
    """
    Plots event map for a given neuron
    
    TODO: finish commenting
    
    """
    
    df[neuron] = df[neuron].where(df[neuron] == 0, 1)
    lw = kwargs.get('linewidth',1)
    s = kwargs.get('s',5) #markersize
    if ax is None:
        figsize = kwargs.get('figsize', (8,8))
        f,ax = plt.subplots(figsize=figsize)
        
    x = df.X_center
    y = df.Y_center

    savefig = kwargs.get('savefig', False)
    filepath = kwargs.get('filepath', None)
    filename = kwargs.get('filename', None)
    title = kwargs.get('title', None)
    
    show_path = kwargs.get('show_path',True)
    a1=0
    if show_path:
        a1 = .1
    ax.scatter(x, y, alpha=a1, linewidth = 0, s=s/2, color = 'grey')
    sns.scatterplot(x = 'X_center', y = 'Y_center', color = 'red', 
                    palette = sns.color_palette(['blue','red']), s = s,
                    marker = 'x', alpha = 0.6, linewidth = lw, hue = behavior,
                    legend = False, hue_order = [0,1], 
                    data = df[df[neuron]>0], ax=ax)
    sns.despine()
    plt.title(title)
    
    if savefig:
        plt.savefig(os.path.join(filepath, filename), dpi = 300)
        
    return ax

def set_weights(x, y, neuron, data, max_rate, framerate=10):
    """Create list of weights by time spent in a location

    Args:
        x: str
            The name of the x-coordinate column in data.

        y: str
            The name of the y-coordinate column in data.

        neuron: str or int
            The name of the neuron column in data.

        data: DataFrame
            The concatenated neuron and behavior DataFrame.

        framerate: int, optional: default: 10
            The framerate of the calcium imaging video.

    Returns:
        weights: list
            A list of values ``w_i`` weighing each coordinate, (x_i, y_i)
    """
    time_spent = data.groupby([x, y]).size()

    # Convert multilevel indexes into columns
    time_spent_df = pd.DataFrame(time_spent)
    time_spent_df.reset_index(inplace=True)

    # Convert coordinate columns and neuron columns into list of 3-tuples
    x_y_count = list(zip(time_spent_df[x], time_spent_df[y], time_spent_df[0]))

    # Convert list of 3-tuples into dict of (x, y): count
    time_at_coords = {(x, y): count for x, y, count in x_y_count}

    neuron = data[neuron].tolist()

    # Convert coordinate columns into list of 2-tuples
    coords_list = list(zip(data[x], data[y]))

    weights = []

    # Go through each component of the neuron column vector, and
    # if the component is not 0, then for the corresponding coordinate-pair,
    # we set the weight to: (flourescence * framerate) divided by the time spent at
    # that location (coordinate-pair).
    for i, coord in enumerate(coords_list):
        if neuron[i] != 0:
            weight = (framerate * neuron[i]) / time_at_coords[coord]
            
            #normalize by the maximum rate 
            weight = weight/max_rate
        else:
            weight = 0

        weights.append(weight)

    return weights

def generate_heatmap(x, y, sigma=2, **kwargs):
    """Generates a heatmap for plotting.

    Wrapper function meant for generating heatmap using the gaussian_filter
    function implemented in scipy, as well as NumPy's histogram2d function.

    Sources:
        https://docs.scipy.org/doc/numpy/reference/generated/numpy.histogram2d.html
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.gaussian_filter.html

    Args:
        x: array_like, shape (N,)
            An array containing the x coordinates of the points to be
            histogrammed.

        y: array_like, shape (N,)
            An array containing the y coordinates of the points to be
            histogrammed.

        bins: int or array_like or [int, int], optional, default: (50, 50)

        sigma: scalar, optional, default: 2
            Standard deviation for Gaussian kernel.

        weights : array_like, shape(N,), optional, default: None
            An array of values ``w_i`` weighing each sample ``(x_i, y_i)``.
            Weights are normalized to 1 if `normed` is True. If `normed` is
            False, the values of the returned histogram are equal to the sum of
            the weights belonging to the samples falling into each bin.

    Returns:
        heatmap: ndarray
            Returned array of same shape as input. We return the transpose of
            the array in order to preserve the view.

        extent: scalars (left, right, bottom, top)
            The bounding box in data coordinates that the image will fill. The
            image is stretched individually along x and y to fill the box.
            Source: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.imshow.html
    """
    if len(x) != len(y):
        raise ValueError("x and y are not of equal length!")

    bins = kwargs.get("bins", (50, 50))
    weights = kwargs.get("weights", None)

    heatmap, xedges, yedges = np.histogram2d(x, y, bins=bins, weights=weights)
    heatmap = gaussian_filter(heatmap, sigma=sigma)
    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    return heatmap.T, extent

def plot_heatmap(x, y, sigma=2, **kwargs):
    """Plots a heatmap.

    Wrapper function for matplotlib and generate_heatmap that plots the actual
    generated heatmap.

    Args:
        x: array_like, shape (N,)
            An array containing the x coordinates of the points to be
            histogrammed.

        y: array_like, shape (N,)
            An array containing the y coordinates of the points to be
            histogrammed.

        bins: int or array_like or [int, int], optional, default: (50, 50)

        sigma: scalar, optional; default: 2
            Standard deviation for Gaussian kernel.

        weights: array_like, shape(N,), optional, default: None
            An array of values ``w_i`` weighing each sample ``(x_i, y_i)``.
            Weights are normalized to 1 if `normed` is True. If `normed` is
            False, the values of the returned histogram are equal to the sum of
            the weights belonging to the samples falling into each bin.

        bounds: tuple, optional, default: None
            If a boundary tuple is provided, the x-axis will be set as follows:
            [bounds[0], bounds[1]]. Similarly, the y-axis will be set as:
            [bounds[0], bounds[1]].

        cmap: matplotlib.colors.LinearSegmentedColormap, optional, default: plt.cm.jet
            The colormap to use for plotting the heatmap.

        figsize: tuple, optional, default: (10, 10)
            The size of the heatmap plot.

        title: str, optional, default: 'Title Goes Here'
            The title of the heatmap plot.
            Note: If title is provided, title will be used as the name of
            the file when the figure is saved.

        dpi: int, optional, default: 600
            The amount of dots per inch to use when saving the figure. In
            accordance with Nature's guidelines, the default is 600.
            Source: https://www.nature.com/nature/for-authors/final-submission

        savefig: bool, optional, default: False
            When True, the plotted heatmap will be saved to the current working
            directory in pdf (IAW Nature's guidelines) format.
            Source: https://www.nature.com/nature/for-authors/final-submission
    """
    if len(x) != len(y):
        raise ValueError("x and y are not of equal length!")

    cmap = kwargs.get("cmap", cm.jet)
    title = kwargs.get("title", None)
    filename = kwargs.get("filename",None)
    filepath = kwargs.get("filepath",None)
    bins = kwargs.get("bins", (50, 50))
    weights = kwargs.get("weights", None)
    figsize = kwargs.get("figsize", (10, 10))
    bounds = kwargs.get("bounds", None)
    x_bounds = kwargs.get("x_bounds", None)
    y_bounds = kwargs.get("y_bounds", None)
    vmin = kwargs.get("vmin",None)
    vmax = kwargs.get("vmax",None)
    ax = kwargs.get('ax',None)
    show_colorbar = kwargs.get('show_colorbar',True)
    # Set user-defined x-axis and y-axis boundaries by appending them to x and y
    if bounds:
        x = x.copy()
        y = y.copy()
        x.loc[len(x)] = bounds[0]
        x.loc[len(x)] = bounds[1]
        y.loc[len(y)] = bounds[0]
        y.loc[len(y)] = bounds[1]
        if not weights is None:
            weights = weights.copy()
            weights.loc[len(weights)] = 0
            weights.loc[len(weights)] = 0
    if x_bounds and y_bounds:
        x = x.copy()
        y = y.copy()
        x.loc[len(x)] = x_bounds[0]
        x.loc[len(x)] = x_bounds[1]
        y.loc[len(y)] = y_bounds[0]
        y.loc[len(y)] = y_bounds[1]
        if not weights is None:
            weights = weights.copy()
            weights.loc[len(weights)] = 0
            weights.loc[len(weights)] = 0


    if ax is None:
        fig,ax = plt.subplots(figsize=figsize)
    heatmap, extent = generate_heatmap(x, y, sigma, bins=bins, weights=weights)
    image = ax.imshow(heatmap, origin="lower", extent=extent, cmap=cmap,vmin=vmin, vmax=vmax)
    if show_colorbar:
        plt.colorbar(image,fraction=.03)
    
    if title:
        plt.title(title)

    dpi = kwargs.get("dpi", 600)
    savefig = kwargs.get("savefig", False)
    if savefig:
        if filename:
            filename = filename
        elif title:
            filename = title
        else:
            filename = "my_smoothed_heatmap"

        plt.savefig("{}{}".format(filepath,filename), dpi=dpi)

##### End of spatial plotting functions #####

def show_traces(mouse, neuron_list, time_period, **kwargs):
    """
    Inputs:
        mouse: mouse object with dataframes for raw, s, and c
        neuron_list: list of neurons to show
        time_period: tuple of ints representing the first and last time bin to show
    """
    m = mouse.name
    b = mouse.session
    fr = mouse.FR
    
    if len(neuron_list) == 0:
        num_plots = len(mouse.spikes.columns)
    else:
        num_plots = len(neuron_list)
        
    figsize = kwargs.get('figsize', (4,3))
    which_traces = kwargs.get('which_traces', ['F','C','S'])
    savefig = kwargs.get('savefig', False)
    filename = kwargs.get('filename', None)
    filepath = kwargs.get('filepath', None)
    dpi = kwargs.get('dpi', 300)
    lw = kwargs.get('linewidth',1)
    fontsize = kwargs.get('fontsize', 6)
    
    colors = kwargs.get('colors',['grey','b','r'])
    
    #if only 1 or 2 colors are given, add the remaining default colors
    n_colors = len(colors)
    if n_colors < 3:
        missing_colors = 3-n_colors
        default_color = ['grey','b','r']
        for i in range(missing_colors):
            colors.append(default_color[i+n_colors])
    
    fig, ax = plt.subplots(num_plots, 1, figsize=figsize)
    title = kwargs.get('title',m+'_'+b)
    

    S = mouse.spikes  # Inferred spikes
    C = mouse.cell_transients  # Denoised fluorescence
    F = mouse.raw  # Raw fluorescence

    if not neuron_list:
        neuron_list = S.columns

    if not time_period:
        #time_period = (0,len(S))
        time_period = (0,S.index[-1])
    t = list(range(time_period[0], time_period[1]))
    x = [j/fr for j in t] #Show seconds, not frames, on the x axis
    
    ylim = kwargs.get('ylim',None)
    for i,n in enumerate(neuron_list):
        
        #Plot in reverse order
        j = (len(neuron_list) - 1) - i
        
        if 'F' in which_traces:
            _=ax[j].plot(x,F.loc[t,n], colors[0],linewidth=lw)
        if 'C' in which_traces:
            _=ax[j].plot(x,C.loc[t,n], colors[1],linewidth=lw)
        if 'S' in which_traces:
            _=ax[j].plot(x,S.loc[t,n], colors[2], linewidth=lw)
        
        ax[j].set_frame_on(False)
        if i!=0:
            ax[j].set_xticks([])
        ax[j].set_yticks([])
        ax[j].set_ylabel(n, rotation=0,fontsize=fontsize)
        
        if ylim:
            ax[j].set_ylim(ylim)
        
    plt.subplots_adjust(wspace=0.0, hspace=0.0)
        

    ax[0].set_title(title, fontsize=fontsize+2, fontweight='bold',fontname='Arial')
    ax[-1].set_xlabel('Time (s)', fontsize = fontsize, fontweight = 'bold', fontname = 'Arial')
    plt.xticks(fontsize = fontsize, fontname = 'Arial')
    
    if savefig:
        plt.tight_layout()
        plt.savefig(os.path.join(filepath, filename), dpi=dpi)
        

def abline(slope, intercept, ax=None, **kwargs):

    """
    Plot a line from slope and intercept
    Inputs:
        slope: float
        intercept: float
    """
    if ax is None:
        fig,ax=plt.subplots()
    x_vals = np.array(ax.get_xlim())
    y_vals = intercept + slope * x_vals
    ax.plot(x_vals, y_vals, **kwargs)
    return ax

    
def pie_chart(sizes, *labels, **kwargs):
    """Wrapper method for matplotlib's pie chart.

    The slices will be ordered and plotted counter-clockwise.

    Args:
        sizes: list
            A list of the sizes of each category.

        labels: str, variable number of args
            The label for each corresponding slice size.

        figsize: tuple, optional, default: (5, 5)
            The size of the figure to be plotted.
    """
    if len(labels) != len(sizes):
        raise ValueError("Length of sizes and amount of labels must be equal.")

    figsize = kwargs.get("figsize", (5, 5))

    _, ax1 = plt.subplots(figsize=figsize)
    ax1.pie(sizes, labels=labels, autopct="%1.1f%%", startangle=90)
    ax1.axis("equal")
    plt.show()

def plot_corr_heatmap(dataframe, **kwargs):
    """Seaborn correlation heatmap wrapper function

    A wrapper function for seaborn to quickly plot a
    correlation heatmap with a lower triangle, only.

    Args:
        dataframe: DataFrame
            A Pandas dataframe to be plotted in the correlation heatmap.

        figsize: tuple, optional, default: (16, 16)
            The size of the heatmap to be plotted.

        title: str, optional, default: None
            The title of the heatmap plot.
            Note: If title is provided, title will be used as the name of
            the file when the figure is saved.

        dpi: int, optional, default: 600
            The amount of dots per inch to use when saving the figure. In
            accordance with Nature's guidelines, the default is 600.
            Source: https://www.nature.com/nature/for-authors/final-submission

        savefig: bool, optional, default: False
            When True, the plotted heatmap will be saved to the current working
            directory in pdf (IAW Nature's guidelines) format.
            Source: https://www.nature.com/nature/for-authors/final-submission
    """
    title = kwargs.get("title", None)

    # Generate a mask for the upper triangle
    mask = np.zeros_like(dataframe.corr(), dtype=np.bool)
    mask[np.triu_indices_from(mask)] = True

    # Set up the matplotlib figure
    _, _ = plt.subplots(figsize=kwargs.get("figsize", (16, 16)))

    # Generate a custom diverging colormap
    cmap = sns.diverging_palette(220, 10, as_cmap=True)

    # Draw the heatmap with the mask and correct aspect ratio
    sns.heatmap(dataframe.corr(), mask=mask, cmap=cmap, vmax=1.0, center=0,
                square=True, linewidths=.5, cbar_kws={"shrink": .5})

    if title:
        plt.title(title)

    dpi = kwargs.get("dpi", 600)
    savefig = kwargs.get("savefig", False)
    if savefig:
        if title:
            filename = title
        else:
            filename = "my_seaborn_heatmap"

        plt.savefig("{}.pdf".format(filename), dpi=dpi)

def plot_clustermap(dataframe, **kwargs):
    """Seaborn clustermap wrapper function

    A wrapper function for seaborn to quickly plot a clustermap using the
    "centroid" method to find clusters.

    Args:
        dataframe: DataFrame
            The Pandas dataframe to be plotted in the clustermap.

        figsize: tuple, optional, default: (15, 15)
            The size of the clustermap to be plotted.

        dendrograms: bool, optional, default: True
            If set to False, the dendrograms (row & col) will NOT be plotted.

        cmap: str, optional, default: "vlag"
            The colormap to use for plotting the clustermap.

        title: str, optional, default: None
            The title of the heatmap plot.
            Note: If title is provided, title will be used as the name of
            the file when the figure is saved.

        dpi: int, optional, default: 600
            The amount of dots per inch to use when saving the figure. In
            accordance with Nature's guidelines, the default is 600.
            Source: https://www.nature.com/nature/for-authors/final-submission

        savefig: bool, optional, default: False
            When True, the plotted heatmap will be saved to the current working
            directory in pdf (IAW Nature's guidelines) format.
            Source: https://www.nature.com/nature/for-authors/final-submission
    """
    cmap = kwargs.get("cmap", "vlag")
    figsize = kwargs.get("figsize", (15, 15))
    title = kwargs.get("title", None)
    dendrograms = kwargs.get("dendrograms", True)

    cluster_map = sns.clustermap(dataframe.corr(), center=0, linewidths=.75, figsize=figsize, method="centroid", cmap=cmap)

    # Set the dendrograms in accordance with passed-in args
    cluster_map.ax_row_dendrogram.set_visible(dendrograms)
    cluster_map.ax_col_dendrogram.set_visible(dendrograms)

    if title:
        cluster_map.fig.suptitle(title)

    dpi = kwargs.get("dpi", 600)
    savefig = kwargs.get("savefig", False)
    if savefig:
        if title:
            filename = title
        else:
            filename = "my_seaborn_clustermap"

        plt.savefig("{}.pdf".format(filename), dpi=dpi)

def plot_contours(A, Cn, thr=None, thr_method='max', maxthr=0.2, nrgthr=0.9,
                  display_numbers=True, max_number=None,
                  cmap=None, swap_dim=False, colors='w', vmin=None,
                  vmax=None, neuron_list = None, show_bg = False, **kwargs):
    """Plots contour of spatial components against a background image
       and returns their coordinates

       From Caiman: https://github.com/flatironinstitute/CaImAn
       @author: agiovann

     Parameters:
     -----------
     A:   np.ndarray or sparse matrix
               Matrix of Spatial components (d x K)

     Cn:  np.ndarray (2D)
               Background image (e.g. mean, correlation)

     thr_method: [optional] string
              Method of thresholding:
                  'max' sets to zero pixels that have value less
                  than a fraction of the max value
                  'nrg' keeps the pixels that contribute up to a
                  specified fraction of the energy

     maxthr: [optional] scalar
                Threshold of max value

     nrgthr: [optional] scalar
                Threshold of energy

     thr: scalar between 0 and 1
               Energy threshold for computing contours (default 0.9)
               Kept for backwards compatibility.
               If not None then thr_method = 'nrg', and nrgthr = thr

     display_number:     Boolean
               Display number of ROIs if checked (default True)

     max_number:    int
               Display the number for only the first max_number components
               (default None, display all numbers)

     cmap:     string
               User specifies the colormap (default None, default colormap)
               
     neuron_list:   list
               User specififies which neurons to plot

     Returns:
     --------
     Coor: list of coordinates with center of mass,
        contour plot coordinates and bounding box for each component
    """
    
    linewidths = kwargs.get('linewidths', 1)
    fontsize = kwargs.get('fontsize', 12)
    fontname = kwargs.get('fontname', 'Arial')
    fontweight = kwargs.get('fontweight', 'bold')
    fontcolor = kwargs.get('fontcolor', colors)
    if sparse.issparse(A):
        A = np.array(A.todense())
    else:
        A = np.array(A)

    if swap_dim:
        Cn = Cn.T
        print('Swapping dim')

    d1, d2 = np.shape(Cn)
    d, nr = np.shape(A)
    if max_number is None:
        max_number = nr

    if neuron_list:
        neuron_list = [i-1 for i in neuron_list]
    else:
        neuron_list = range(np.minimum(nr, max_number))
        
    if thr is not None:
        thr_method = 'nrg'
        nrgthr = thr
        print("The way to call utilities.plot_contours has changed.")

    if show_bg:    
        x, y = np.mgrid[0:d1:1, 0:d2:1]
    else:
        x, y = np.mgrid[d1:0:-1, 0:d2:1] #reverse y direction to get correct orientation
    ax = plt.gca()
    
    if show_bg:
        if vmax is None and vmin is None:
            plt.imshow(Cn, interpolation=None, cmap=cmap,
                       vmin=np.percentile(Cn[~np.isnan(Cn)], 1),
                       vmax=np.percentile(Cn[~np.isnan(Cn)], 99))
        else:
            plt.imshow(Cn, interpolation=None, cmap=cmap,
                       vmin=vmin, vmax=vmax)
   
    coordinates = []
    cm = au.com(A, d1, d2)
    for i in neuron_list:
        pars = dict(kwargs)
        if thr_method == 'nrg':
            indx = np.argsort(A[:, i], axis=None)[::-1]
            cumEn = np.cumsum(A[:, i].flatten()[indx]**2)
            cumEn /= cumEn[-1]
            Bvec = np.zeros(d)
            Bvec[indx] = cumEn
            thr = nrgthr

        else:
            if thr_method != 'max':
                print("Unknown threshold method. Choosing max")
            Bvec = A[:, i].flatten()
            Bvec /= np.max(Bvec)
            thr = maxthr

        if swap_dim:
            Bmat = np.reshape(Bvec, np.shape(Cn), order='C')
        else:
            Bmat = np.reshape(Bvec, np.shape(Cn), order='F')
        cs = plt.contour(y, x, Bmat, [thr], colors=colors, linewidths = linewidths)

        # this fix is necessary for having disjoint figures and borders
        p = cs.collections[0].get_paths()
        v = np.atleast_2d([np.nan, np.nan])
        for pths in p:
            vtx = pths.vertices
            num_close_coords = np.sum(np.isclose(vtx[0, :], vtx[-1, :]))
            if num_close_coords < 2:
                if num_close_coords == 0:
                    # case angle
                    newpt = np.round(au.old_div(vtx[-1, :], [d2, d1])) * [d2, d1]
                    vtx = np.concatenate((vtx, newpt[np.newaxis, :]), axis=0)

                else:
                    # case one is border
                    vtx = np.concatenate((vtx, vtx[0, np.newaxis]), axis=0)

            v = np.concatenate((v, vtx, np.atleast_2d([np.nan, np.nan])),
                               axis=0)

        pars['CoM'] = np.squeeze(cm[i, :])
        pars['coordinates'] = v
        pars['bbox'] = [np.floor(np.min(v[:, 1])), np.ceil(np.max(v[:, 1])),
                        np.floor(np.min(v[:, 0])), np.ceil(np.max(v[:, 0]))]
        pars['neuron_id'] = i + 1
        coordinates.append(pars)

    if display_numbers:
        for i in neuron_list:
            
            x = cm[i,1]
            if show_bg: 
                y = cm[i,0]
            else:
                y = d1-cm[i,0] #need to flip to preserve location if theres no background
                
            if swap_dim:
                ax.text(y, x, str(i + 1), color=colors)
            else:
                ax.text(x, y, str(i + 1), color=fontcolor, fontsize = fontsize, fontname = fontname, fontweight = fontweight) 

    return coordinates

def raster_plot(mouse, neurons=[], binsize=2, behaviors=[], colors=[], ax=None):
    """
    Plots summary of neural data as a raster, and as a histogram showing a sum
    of the total activity per time bin
    
    Inputs:
        
        mouse: Mouse object
        neurons: list of ints (optional)
            List of the neurons to plot. Default is to show all neurons
        binsize: int
            Length of each time bin, in seconds. Applies to lower plot only.
        behaviors: list of str (optional)
            Names of behavior columns in mouse.behavior to display.
        colors: list
            Colors of behaviors
            
    Returns:
        ax: matplotlib.axes object
    """
    
    if ax is None:
        fig,ax = plt.subplots(2,1,figsize=(2.5,3))

    current_FR = mouse.FR #in frames per second
    framelength = 1/current_FR
    str_fr = str(int(1000/current_FR))+'ms' 
    bin_fr = str(binsize)+'s'

    s=mouse.spikes
    
    if len(neurons)==0:
        neurons = s.columns
    
    xax = [framelength*i for i in s.index]
    events=[]    
    for i,col in enumerate(neurons):
        n = s[col]
        times = n[n>0].index.tolist()

        #convert times from frame number to time
        times = [t*framelength for t in times]

        events.append(times)
    ax[0].eventplot(events, color='k',linelengths=.5, linewidths=.5)

    s = au.resample_ts(s[neurons], current_FR = str_fr, new_FR=bin_fr,method='sum')
    spikes_sum = s.mean(axis=1)

    xax_bin = [binsize*i for i in spikes_sum.index]
        
    ax[1].bar(xax_bin, spikes_sum.values, width=binsize, color='k',linewidth=1,edgecolor='k')

    if len(mouse.behavior)>0:
        for index, b in enumerate(behaviors):
            try:
                evt=au.extract_epochs(mouse,b)
                interval = au.filter_epochs(evt[1],framerate=mouse.FR, seconds=.2)
                alpha=.5

                color = colors[index]

                for i in interval:
                    # convert i from frame number to seconds
                    j = [t*framelength for t in i]
                    ax[0].axvspan(j[0], j[-1], alpha=alpha, color=color,linewidth=0)
                    ax[1].axvspan(j[0], j[-1], alpha=alpha, color=color,linewidth=0)
            except:
                continue
    ax[1].set_xticks(ax[1].get_xticks(),ax[0].get_xticklabels())
    ax[0].set_xlim(0,np.max(xax))
    ax[1].set_xlim(0,np.max(xax))
    #ax[1].set_ylim(0,.3)
    ax[0].set_xticks([])
    ax[0].set_title(mouse.name)
    ax[0].set_ylabel('Cell Number')
    ax[1].set_ylabel('Activity')
    ax[1].set_xlabel('Time (s)')
    ax[0].axes.spines['bottom'].set_visible(False)
    plt.subplots_adjust(wspace=0, hspace=0)
    return ax

def cum_dist_plot(var=None, hue=None, data=None, ax=None, **kwargs):
    if ax is None:
        figsize = kwargs.get('figsize',(2,2))
        fig,ax = plt.subplots(figsize=figsize)

    hue_vals = kwargs.get('order',list(set(data[hue])))

    colors = kwargs.get('colors',None)
    cmap = kwargs.get('cmap','Reds')
    colors = kwargs.get('colors',None)
    lw = kwargs.get('linewidth',1)
    
    colormap=None
    if colors is None:     
        colormap = get_cmap(len(hue_vals),cmap)
    
    #find max and min values and draw horizontal line for each hue to extend
    max_val = np.max([data[data[hue]==h][var].max() for i,h in enumerate(hue_vals)])
    min_val = np.min([data[data[hue]==h][var].min() for i,h in enumerate(hue_vals)])
    
    for i,h in enumerate(hue_vals):
        a=np.sort(data[data[hue]==h][var])
        label=kwargs.get('label',h)
        if colormap:
            c=colormap(i)
        else:
            c=colors[i]   
        ax.step(a,np.arange(len(a))/len(a),color=c, label=label,linewidth=lw)
        
        #draw lines to ensure all hues extend to the same x coordinate
        
        
        ax.hlines(y=1, xmin=np.max(a), xmax=max_val, color=c, linewidth=lw)
        ax.hlines(y=0, xmin=min_val, xmax=np.min(a), color=c, linewidth=lw)
        
        #draw vertical line to ensure data lines go to y=1
        
        y_max_data = 1-1/len(a)
        ax.vlines(x=np.max(a), ymin=y_max_data, ymax=1, color=c, linewidth=lw)
    title= kwargs.get('title',None)
    if title:
        ax.set_title(title)
    ax.set_xlabel(var)
    ax.set_ylabel('Cumulative \nProportion')
    return ax

def get_cmap(n, name="tab20"):
    """ Returns a function that maps each index in 0,1,...,n-1 
        to a distinct RGB color; the keyword argument name must 
        be a standard mpl colormap name.
    """
    return plt.cm.get_cmap(name, n)

def plot_positions_dlc(rois, tracking=None,xlim=[5,85], ylim=[-5,65], style='line',**kwargs):
    from matplotlib.patches import Polygon
    #from matplotlib.collections import PatchCollection
    
    bins = kwargs.get('bins',(50,50))
    sigma = kwargs.get('sigma',2)
    cmap=kwargs.get('cmap','Greys')
    autofit = kwargs.get('autofit',True)
    which_rois = kwargs.get('which_rois',None)
    roi_colors = kwargs.get('roi_colors',['k','b','r'])
    title = kwargs.get('title',None)
    roi_linewidth = kwargs.get('linewidth',7)
    pos_linewidth = kwargs.get('pos_lw',1)
    ax=kwargs.get('ax',None)
    
    if ax is None:
        fig,ax = plt.subplots(figsize=(10,10))
    patches=[]

    for i,r in enumerate(rois):
        if which_rois:
            if i not in which_rois:
                continue;
        patches.append(Polygon(r[1], edgecolor = roi_colors[i], linewidth=roi_linewidth, fill=False))
                
    for p in patches:
        ax.add_patch(p)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim[1],ylim[0]) #flip y axis
    ax.set_xticks([])
    ax.set_yticks([])
    spines = ['top','bottom','left','right']
    for s in spines:
        ax.spines[s].set_visible(False)
    ax.set_aspect('equal')
    ax.set_title(title)
    if tracking is not None:      
        if style=='line':
            ax.plot(tracking['x'], tracking['y'], color = 'grey', alpha=.5,linewidth=pos_linewidth)
            
        elif style=='heatmap':
            import analysis.visualize as vis
            show_colorbar = kwargs.get('show_colorbar',False)
            weights = pd.Series(np.tile(1,len(tracking)))
            if autofit:
                arena_x = list(zip(*rois[0][1]))[0]
                arena_y = list(zip(*rois[0][1]))[1]
                x_bounds = (np.min(arena_x), np.max(arena_x))
                y_bounds = (np.min(arena_y), np.max(arena_y))
                vis.plot_heatmap(tracking['x'], tracking['y'], sigma=sigma, bins=bins, ax=ax, show_colorbar=show_colorbar,
                             x_bounds=x_bounds, y_bounds=y_bounds, cmap=cmap,vmin=.01,norm=True,weights=weights)         
            else:
                vis.plot_heatmap(tracking['x'], tracking['y'], sigma=sigma, bins=bins, ax=ax, show_colorbar=show_colorbar,
                            cmap=cmap,vmin=.01,norm=True,weights=weights)
        else:
            print('Only plot styles of "line" or "heatmap" are supported')
    
    return ax

# Plots for Single Neuron Properties (e.g. selectivity)

def selectivity_stacked_bar(selectivity_counts_df, colors):
    ind = np.arange(selectivity_counts_df.shape[1])
    width = 0.75
    
    darker = [adjust_lightness(color, amount=0.8) for color in colors]
    lighter = [adjust_lightness(color, amount=1.5) for color in darker]
    
    act = selectivity_counts_df.loc['activated',:]
    inhib = selectivity_counts_df.loc['inhibited',:]
    ns = selectivity_counts_df.loc['non_selective',:]
    
    fig, ax = plt.subplots()
    ax.bar(ind, act, width, color=darker)
    ax.bar(ind, inhib, width, bottom=act, color=lighter)
    ax.bar(ind, ns, width, bottom=act+inhib, color='lightgray')
    for i in range(len(act)):
        ax.text(ind[i], act[i]/2, str(int(np.round(act[i]*100)))+'%\nact.', fontsize=3, ha='center', va='center', color='black')
        ax.text(ind[i], inhib[i]/2 + act[i], str(int(np.round(inhib[i]*100)))+'%\ninh.', fontsize=3, ha='center', va='center', color='black')
        ax.text(ind[i], ns[i]/2 + act[i] + inhib[i], str(int(np.round(ns[i]*100)))+'%\nns', fontsize=3, ha='center', va='center', color='black')
    ax.set_xticks(ind)
    ax.set_xticklabels(selectivity_counts_df.columns, rotation=25)
    #ax.axes.get_yaxis().set_visible(False)
    ax.set_title('Selective Neuron Proportions')
    return ax

def SelectivityOverlapHeatmap(p_value_df, odds_ratio_df, p_cutoff=0.05, save=False, save_path=None):
    
    log_df = -np.log(p_value_df) # Generate a -log(p-value) DataFrame from the p-value DataFrame
    
    # Set up a MultiIndex IndexSlice object and define the categories for cell-selectivity
    idx = pd.IndexSlice
    selectivity = ['activated', 'inhibited', 'non_selective']
    
    # Generate a heatmap for each pair of cell-selectivity categories and highlight the cells which have a p-value
    # that is smaller than the defined p-cutoff, pairs which are overlapping are highlighted lightcoral and pairs
    # that are distinct are highlighted cornflowerblue
    for pair in list(combinations_with_replacement(selectivity, 2)):
        
        # Subset the -log(p-value) Dataframe
        log_subset = log_df.loc[idx[:,pair[0]], idx[:,pair[1]]]
        log_subset.index = log_subset.index.droplevel(1)
        log_subset.columns = log_subset.columns.droplevel(1)
        
        # Generate an upper-triangular mask to hide the upper portion of the heatmap
        mask = np.ones_like(log_subset, dtype=np.bool)
        mask[np.triu_indices_from(mask)] = False
        
        # Plot the -log(p-value) subset as a heatmap
        ax = sns.heatmap(data=log_subset * mask, square=True, cmap='Greys', cbar_kws={'shrink':0.5, 
                                                                                      'label':'-log(p value)'})

        # Get the integer indices for the entries which have a -log(p-value) that is greater than the -log(p-cutoff)
        x, y = np.where(log_subset >= -np.log(p_cutoff))
        coord = [(i,j) for i,j in zip(x,y) if i < j]
        
        # Subset the odds ratio DataFrame
        odds_subset = odds_ratio_df.loc[idx[:,pair[0]], idx[:,pair[1]]]
        odds_subset.index = odds_subset.index.droplevel(1)
        odds_subset.columns = odds_subset.columns.droplevel(1)
        
        # Highlight the cells which correspond to cell-selectivity pairs that have p-values below the p-cutoff
        for co in coord:
            if odds_subset.iloc[co] >= 1:
                ax.add_patch(Rectangle(co, 1, 1, fill=False, edgecolor='lightcoral', lw=1))
            elif odds_subset.iloc[co] < 1:
                ax.add_patch(Rectangle(co, 1, 1, fill=False, edgecolor='cornflowerblue', lw=1))
                
        plt.xlabel(pair[1])
        plt.ylabel(pair[0])
        plt.title('Cell Selectivity Overlap')
        
        if save == True:
            plt.savefig(save_path + f'{pair[0]}_{pair[1]}_overlap_heatmap.png')
            
        plt.show()

        # Helper Functions

def adjust_lightness(color, amount=0.5):
    """
    Changes the lightness of the given color by multiplying by the given amount.
    Input can be matplotlib color string, hex string, or RGB tuple. Colors get
    lighter with amount > 1 and darker with amount < 1.
    
    Source: 
    https://stackoverflow.com/questions/37765197/darken-or-lighten-a-color-in-matplotlib

    Examples:
    >> adjust_lightness('g', 0.3)
    >> adjust_lightness('#F034A3', 0.6)
    >> adjust_lightness((.3,.55,.1), 0.5)
    """
    import matplotlib.colors as mc
    import colorsys
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


import math
import pandas as pd
import analysis.analysis_utils as au
import copy

def pair(x, y):
    z = (((x + y + 1) * (x + y)) / 2) + y
    return z

def invert(z):
    w = math.floor(((math.sqrt(8*z + 1) - 1) / 2))
    t = (w**2 + w) / 2
    y = z - t
    x = w - y
    return x, y

def apply_cantor_pairing(x_coords, y_coords):
    """Reduce dimensionality from 2d to 1d.

    Args:
        x_coords: list

        y:coords: list

    Returns:
        z_coords: list
    """
    if len(x_coords) != len(y_coords):
        raise ValueError("The length of x_coords must equal the length of y_coords!")

    z_coords = []
    for x, y in zip(x_coords, y_coords):
        z = pair(x, y)
        z_coords.append(z)

    return z_coords


def bin_cantor_spatial(mouse, bin_size=5):
    """
    Bins the x and y coordinates, and transforms the new x,y cords to z, to be used for 
    spatial information computations

    Inputs
        mouse: Mouse object
        bin_size: bin size, in cm. Default is 5
        
    Returns:
        x_coords: Series, binned x coordinates
        y_coords: Series, binned y coordinates
        z_coords: Series, unique labels for (x,y) bins
    """
   
    # Extract the X and Y coordinate column vectors
    x_coords = mouse.spikes_and_beh["X_center"].astype(float)
    y_coords = mouse.spikes_and_beh["Y_center"].astype(float)

    # Subtract the mean value of the coordinate to ensure that the binning centers around
    # the main track
    x_coords -= x_coords.mean() 
    y_coords -= y_coords.mean()

    ## Bin X and Y coordinates by bin_factor (ie bin_factor=5 -> 5cm x 5cm bins)
    bin_factor=bin_size
    x_coords = (x_coords/bin_factor).astype(int)
    y_coords = (y_coords/bin_factor).astype(int)


    # Shift all coordinate values by increasing all of them by the minimum value. This
    # is necessary in order to apply the cantor pairing function, since the cantor 
    # pairing function is only defined on the natural numbers, i.e., {0, 1, 2, 3, ...}.
    x_coords += abs(x_coords.min())
    y_coords += abs(y_coords.min())

    #plt.scatter(x_coords,y_coords)
    
    # Reduce the dimensionality of the coordinates, since sklearn's mutual information 
    # function only allows you to compute the MI between two arrays.
    z_coords = apply_cantor_pairing(x_coords.tolist(), y_coords.tolist())
    z_coords = pd.Series(data=z_coords)
    
    return x_coords,y_coords,z_coords

def modify_mouse(mouse, spatial_bin, min_time):
    """
    Modifies a mouse object so that it is ready for spatial/place cell analysis.
    Specifically, immobile times bins are removed, the spatial positions are 
    binned, and spatial bins with low occupancy are removed. The return from this function
    is a mouse object that can be used to compute spatial information, and plot
    spatial heatmaps using functions in the visualize module.
    
    Inputs:
        mouse: Mouse object
            The mouse object to be modified
        spatial_bin: int
            Size of the spatial bins, in cm
        min_time: int
            The minimum number of frames that a mouse can occupy a spatial bin
    
    """
    
    mobile_mouse = remove_immobile(mouse)
    #mobile_mouse = copy.deepcopy(mouse)
    x,y,z = bin_cantor_spatial(mobile_mouse, bin_size = spatial_bin)    
    modified_mouse = remove_low_occupancy(mobile_mouse, x, y, min_occupancy=min_time)    
    
    return modified_mouse

def remove_immobile(mouse):
    """
    Removes immobile time bins from mouse. Assumes that mouse.behavior has an "immobile" column.
    Intended to be used as a helper function for modify_mouse
    
    Inputs:
        mouse: Mouse object
    
    Return:
        Mouse object with immobile time bins removed
    """
    
    
    mobile_S = mouse.spikes.loc[mouse.behavior.immobile == 0]
    mobile_beh = mouse.behavior[mouse.behavior.immobile == 0]
    mobile_C = mouse.cell_transients.loc[mouse.behavior.immobile == 0]
    mobile_raw = mouse.raw.loc[mouse.behavior.immobile == 0]
    mobile_sb = mouse.spikes_and_beh[mouse.behavior.immobile == 0]
    #mobile_sb = mobile_S.join(mobile_beh,how='left')
    
    new_mouse = copy.deepcopy(mouse)
    new_mouse.spikes = mobile_S
    new_mouse.behavior = mobile_beh
    new_mouse.cell_transients = mobile_C
    new_mouse.spikes_and_beh = mobile_sb
    new_mouse.raw = mobile_raw
    
    return new_mouse

def remove_low_occupancy(mouse, x_bin, y_bin, min_occupancy = 1):
    """
    Removes spatial bins that were not visited much. 
    Intended to be used as a helper function for modify_mouse

    
    Inputs:
        mouse: Mouse object upon which to modify
        x_bin: Series of binned x coordinates
        y_bin: Series of binned y coordinates
        min_occupancy: minimum number of time bins for a spatial bin to be included
        
    Returns:
        Mouse object with low-occupancy spatial bins removed, and 
        the new columns (X_bin, Y_bin, Z) in the behavior and 
        spikes_and_beh dataframes     
    """
    
    z_coords = apply_cantor_pairing(x_bin.tolist(), y_bin.tolist())
    z_coords = pd.Series(data=z_coords)
    
    df = mouse.spikes_and_beh
    df.loc[:,'X_bin'] = x_bin
    df.loc[:,'Y_bin'] = y_bin
    df.loc[:,'Z'] = z_coords
    df.loc[:,'time_bin'] = df.index
    df.set_index(['X_bin','Y_bin'],inplace=True)
    occupancy = df.groupby(['X_bin','Y_bin']).size().to_frame()
    occupancy.rename({0:'occupancy'},axis=1,inplace=True)

    df_merge = df.merge(occupancy,left_index=True,right_index=True)
    df_merge.reset_index(inplace=True)
    df_merge.set_index('time_bin',inplace=True)
    df_merge.sort_index(inplace=True)

    df.reset_index(inplace=True)
    df.set_index('time_bin',inplace=True)
    df.sort_index(inplace=True)
    df
    df_merge
    df_filtered = df_merge[df_merge.iloc[:,-1]>=min_occupancy]

    z_coords = apply_cantor_pairing(df_filtered.X_bin.tolist(), df_filtered.Y_bin.tolist())
    z_coords = pd.Series(data=z_coords, index = df_filtered.index)
    
    new_mouse = copy.deepcopy(mouse)
    
    new_mouse.raw = mouse.raw[mouse.raw.index.isin(df_filtered.index)]
    new_mouse.spikes = mouse.spikes[mouse.spikes.index.isin(df_filtered.index)]
    filtered_binned_beh = mouse.behavior[mouse.behavior.index.isin(df_filtered.index)]
    filtered_binned_beh.loc[:,'Z'] = z_coords
    filtered_binned_beh.loc[:,'X_bin'] = df_filtered.X_bin
    filtered_binned_beh.loc[:,'Y_bin'] = df_filtered.Y_bin
    new_mouse.behavior=filtered_binned_beh
    
    new_mouse.cell_transients = mouse.cell_transients[mouse.cell_transients.index.isin(df_filtered.index)]
    new_mouse.spikes_and_beh = new_mouse.spikes.join(new_mouse.behavior, how='left')
    
    return new_mouse



"""
This module provides functions for drawing regions of interest on a video, and determining whether
an (x,y) point lies within that ROI.

@author: Sean Lim, selimchl@gmail.com
May 2, 2019
"""

from shapely.geometry import Point, Polygon
import numpy as np
import holoviews as hv
from holoviews import streams
import cv2
#from holoviews import opts
#from holoviews.streams import Stream, param
hv.notebook_extension('bokeh')



def draw_rois(vid_path, num_roi, invert_yaxis=True, draw_frame=0):
    """
    Displays a frame of a given video, and promps the user to draw ROIs on the image

    Inputs:
        vid_path: str
            Path to the desired video
        num_roi: int
            Number of ROIs
        invert_yaxis: bool, optional
            Whether to invert the y axis when displaying the video frame.
            Defauls is True
        draw_frame: int, optional
            The frame to display. Default is 0 (first frame)
    Returns:
        image: holoviews image
        roi_stream: PolyDraw object containing roi data

    """
    vid = cv2.VideoCapture(vid_path)

    #set the first frame of the video to frame
    _ = vid.set(cv2.CAP_PROP_POS_FRAMES, draw_frame)
    _, frame = vid.read()
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    vid.release()
    
    image = hv.Image((np.arange(frame.shape[1]), np.arange(frame.shape[0]), frame))
    image.opts(width=int(frame.shape[1]),
               height=int(frame.shape[0]),
               invert_yaxis=invert_yaxis, cmap='gray',
               colorbar=False,
               toolbar='below',
               title="Draw ROIs in Name Order")
    
    #draw the rois with holoviews PolyDraw tool
    roi = hv.Polygons([])
    roi.opts(alpha=0.5)
    roi_stream = streams.PolyDraw(source=roi, drag=True, num_objects=num_roi)

    image = image * roi

    return image, roi_stream

def draw_rois_frame(frame_path, num_roi, invert_yaxis=True, draw_frame=0):
    """
    Uses an extracted frame from a video, and promps the user to draw ROIs on the image

    Inputs:
        frame_path: str
            Path to the extracted frame
        num_roi: int
            Number of ROIs
        invert_yaxis: bool, optional
            Whether to invert the y axis when displaying the video frame.
            Defauls is True
        draw_frame: int, optional
            The frame to display. Default is 0 (first frame)
    Returns:
        image: holoviews image
        roi_stream: PolyDraw object containing roi data

    """
    frame = frame_path
    
    image = hv.Image((np.arange(frame.shape[1]), np.arange(frame.shape[0]), frame))
    image.opts(width=int(frame.shape[1]),
               height=int(frame.shape[0]),
               invert_yaxis=invert_yaxis, cmap='gray',
               colorbar=False,
               toolbar='below',
               title="Draw ROIs in Name Order")
    
    #draw the rois with holoviews PolyDraw tool
    roi = hv.Polygons([])
    roi.opts(alpha=0.5)
    roi_stream = streams.PolyDraw(source=roi, drag=True, num_objects=num_roi)

    image = image * roi

    return image, roi_stream


def get_rois(roi_stream, names, num_roi):
    """
    Get the coordinates of the vertices for the drawn rois

    Inputs:
        roi_stream: holoviews PolyDraw
            Contains ROI position data
        names: list
            Names of ROI's
        num_roi: int
            number of ROI's
    Returns:
        rois: list of tuples
            (x,y) coordincates corresponding to each ROI
    """

    rois = []
    for iters in range(0, num_roi):
        roi_coord = list(zip(roi_stream.data['xs'][iters], roi_stream.data['ys'][iters]))
        rois.append(tuple([names[iters], roi_coord]))
    return rois

def point_in_zone(points, roi):
    """
    Intended to be used as a helper function for in_zone()

    Inputs:
        points: list of shapely Point objects
        roi: shapely Polygon object

    Returns:
        bool_vec: numpy array
            Boolean vector representing whether the (x,y) point
            is in the ROi or not
    """

    bool_vec = []
    for value in points:
        bool_vec.append(int(value.within(roi)))
    return np.array(bool_vec)


def in_zone(x_pos, y_pos, rois):
    """
    Takes in an x-position array and a y-position array and returns a boolean
    array that tells you if the point is
    in the roi or not

    Inputs:
        x_pos: T x 1 numpy array or pandas Series
            x position at all timepoints
        y_pos: T x 1 numpy array or pandas Series
            y position at all timepoints
        rois: list of tuples
            output of getROIs(), the coordinates defining the ROIs
    Returns:
        points_bool: list of tuples
            list of (name, boolean vector) tuples defining whether the point
            is in the ROI for each time point
    """

    if len(x_pos) != len(y_pos):
        raise ValueError('x and y position arrays are not the same length')
    else:
        #convert the rois to shapely Polygon objects
        polys = []
        for value in rois:
            polys.append(tuple([value[0], Polygon(value[1])]))

        #convert the x and y position arrays to a list of shapely Point objects
        points = []
        point_tuple = tuple(zip(x_pos, y_pos))
        for value in point_tuple:
            points.append(Point(value))

        #calc boolean vector for each roi
        points_bool = []
        for poly in polys:
            point_bool = point_in_zone(points, poly[1])
            points_bool.append(tuple([poly[0], point_bool]))

    return points_bool

def set_coord(roi, points, x=None, y=None):
    """
    Sets x or y positions of a list of roi points to a new value.

    Inputs:
        roi: list of tuples
            List of rois in the format: [(x_0, y_0)...(x_n, y_n)]
        points: list
            List of the indices of rois to modify (eg. points = [0] will modify roi[0])
        x: float, optional
            New x value to set
        y: float, optional
            New y value to set
    """
    for point in points:
        if x is None:
            x1 = roi[point][0]
        else:
            x1 = x
        if y is None:
            y1 = roi[point][1]
        else:
            y1 = y
        roi[point] = (x1, y1)
        


"""
This module contains all the functions necessary for data preprocessing as well
as data wrangling.

@authors: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
          Jack Berry, Columbia University, jeb2242@columbia.edu
"""

import os
import sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from analysis.resampling import Resampler

from itertools import product
from scipy.stats import fisher_exact 

def read_ethovision(etho_filepath):
    """
    Reads in data from an Ethovision raw data file (converted to .csv format).
    Expected format is a T x B matrix, where T is time bins, and B is the
    automatically tracked behaviors.

    Intended to to be used as a helper function for preprocess_behavior

    Args:
        etho_filepath: str
            The file path to the raw Ethovision .csv file

    Returns:
        behavior: pandas DataFrame
            The processed Ethovision data
    """

    # read ethovision data and rename columns
    nrowsheader = pd.read_csv(etho_filepath, nrows=0, encoding="latin1", header=0)
    nrowsheader = int(nrowsheader.columns[1])

    behavior = pd.read_csv(etho_filepath, header=nrowsheader-2, encoding="latin1", low_memory=False)
    behavior = behavior.drop(behavior.index[0])
    behavior = behavior.apply(pd.to_numeric, errors='coerce')

    # rename columns
    new_cols = []
    for col in behavior.columns:
        col = col.replace(' ', '_').replace("In_zone(", '').replace("_/_center-point)", '')
        new_cols.append(col)

    behavior.columns = new_cols

    # convert Trial_time to timedelta function to enable time operations
    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit='s')

    return behavior

def read_observer(observer_filepath):
    """
    Reads in data from an observer file (converted to .csv format). Expected
    format is a T x B matrix, where T is time bins, and B is the hand-scored
    behaviors.

    Intended to to be used as a helper function for preprocess_behavior

    Args:
        observer_filepath: str
            The file path to the raw observer .csv file

    Returns:
        behavior: pandas DataFrame
            The processed observer data

    """
    behavior = pd.read_csv(observer_filepath)
    behavior = behavior.drop(["Observation", "Event Log", "Time"], axis=1)
    behavior.index += 1
    behavior.fillna(0, inplace=True)
    behavior = behavior.apply(pd.to_numeric, errors='coerce')
    behavior[behavior != 0] = 1

    return behavior

def preprocess_behavior(etho_filepath=None, observer_filepath=None):
    """
    Processes original ethovision and observer files (converted to .csv format)

    Args:
        etho_filepath: str
            The file path to raw ethovision behavior .csv file.

        observer_filepath: str
            The file path to the raw observer .csv file.

    Returns:
        behavior: DataFrame
            The preprocessed behavior DataFrame.
    """

    if etho_filepath and observer_filepath:
        #Read in behavior files
        etho_behavior = read_ethovision(etho_filepath)
        obs_behavior = read_observer(observer_filepath)
        num_obs = len(obs_behavior.columns)
        
        #Merge Ethovision and Observer data
        behavior = pd.merge(etho_behavior, obs_behavior, how="left", left_index=True, right_index=True)
        behavior.update(behavior[behavior.columns[-num_obs:]].fillna(0))
        
        return behavior

    elif etho_filepath:
        etho_behavior = read_ethovision(etho_filepath)
        return etho_behavior
        
    elif observer_filepath:
        obs_behavior = read_observer(observer_filepath)
        return obs_behavior
    
    else:
        raise ValueError("You did not provide an ethovision or observer file path")

def z_score_data(data, mew=None):
    """This function simply z scores all the given neural signal data.

    Args:
        data: DataFrame
            Ca2 transient data in T x N format, where N is the # of neuron
            column vectors, and T is the number of observations (rows),
            all in raw format.

        mew: int, optional, default: None
            The mean value for the baseline of the data.
            Note: if the raw data comes from CNMF_E, use 0 for the baseline.

    Returns:
        z_scored_data: DataFrame
            The z scored raw cell data in T x N format.
    """
    pop_offset = np.percentile(data.values, 50)
    sigma = np.std(data.values, axis=0)

    if mew is None:
        mew = np.mean(data.values[data.values < pop_offset])
    else:
        mew = mew

    z_scored_data = pd.DataFrame(data=((data.values - mew) / sigma))
    return z_scored_data


def pairwise_dist(x_coords):
    """Computes consecutive pairwise differences in a Series.

    NOTE: This is a helper function for distance_moved.

    Args:
        x_coords: pandas Series
            A one-dimensional ndarray of x coordinates, over time.

    Returns:
        dx: pandas Series
            A series containing all the values x[i] - x[i-1]

    """
    x_coords = x_coords.astype(float)
    delta_x = (np.roll(x_coords, -1) - x_coords).shift(1)
    return delta_x


def distance_moved(x_coords, y_coords):
    """Computes the distance moved per frame

    Args:
        x_coords: pandas Series
            A one-dimensional ndarray of x coordinates, over time.

        y_coords:
            A one-dimensional ndarray of y coordinates, over time.

    Returns
        Series containing distance moved per frame

    """
    if len(x_coords) != len(y_coords):
        raise ValueError("x_coords and y_coords are not of equal length!")

    delta_x = pairwise_dist(x_coords.astype(float))
    delta_y = pairwise_dist(y_coords.astype(float))

    dist_moved = delta_x**2 + delta_y**2
    dist_moved = dist_moved.apply(np.sqrt)
    return dist_moved


def compute_velocity(dist_moved, framerate=10):
    """
    Args:
        dist_moved: pandas Series
            A one-dimensional ndarray containing distance moved.

        framerate: int, optional, default: 10
            The frame rate corresponding to the dist_moved Series.

    Returns:
        velocity: pandas Series
            A one-dimensional ndarray containing velocity.
    """
    velocity = dist_moved.apply(lambda x: x * framerate)
    return velocity


def define_immobility(velocity, min_dur=1, min_vel=2, framerate=10, min_periods=1):
    """Define time periods of immobility based on a rolling window of velocity.

    A Mouse is considered immobile if velocity has not exceeded min_vel for the
    previous min_dur seconds.

    Default values for min_dur and min_vel are taken from:
    Stefanini...Fusi et al. 2018 (https://doi.org/10.1101/292953)

    Args:
        velocity: pandas Series
            A one-dimensional ndarray of the velocity data.

        min_dur: int, optional, default: 1
            The minimum length of time in seconds in which velocity must be low.

        min_vel: int, optional, default: 2
            The minimum velocity in cm/s for which the mouse can be considered
            mobile.

        framerate: int, optional, default: 10
            The frame rate of the velocity Series. Default is 10 fps.

        min_periods: int, optional, default: 1
            Minimum number of datapoints needed to determine immobility. This
            value is needed to define immobile time bins at the beginning of the
            session. If min_periods=8, then the first 8 time bins will be be
            considered immobile, regardless of velocity.

    Returns:
        mobile_immobile: pandas Series
            A one-dimensional ndarray of 0's and 1's, where 1 signifies immobile
            times and 0 signifies mobile times.

    """
    window_size = framerate * min_dur
    rolling_max_vel = velocity.rolling(window_size, min_periods=min_periods).max()
    mobile_immobile = (rolling_max_vel < min_vel).astype(int)

    return mobile_immobile


def find_file(root_directory, target_file):
    """Finds a file in a given root directory (folder).

    Args:
        root_directory: str
            The name of the first or top-most directory (folder)
            to search for the target file (e.g. "Hen_Lab/Mice").

        target_file: str
            The full name of the file to be found (e.g. "mouse1_spikes.csv").

    Returns:
        file_path: str
            The full path to the target file.

    """
    root_directory = os.path.join(os.path.expanduser("~"), root_directory)

    if not os.path.exists(root_directory):
        raise FileNotFoundError("{} does not exist!".format(root_directory))
    if not os.path.isdir(root_directory):
        raise NotADirectoryError("{} is not a directory!".format(root_directory))

    for subdir, _, files in os.walk(root_directory):
        for file in files:
            if file == target_file:
                file_path = os.path.join(subdir, file)
                return file_path

    raise FileNotFoundError("{} not found!".format(target_file))


def extract_epochs(mouse, behavior):
    """Extract all epochs of a continuous behavior/event.

    Args:
        mouse: Mouse
            An instance of the Mouse class that has a `spikes_and_beh`
            pandas DataFrame, where `spikes_and_beh` has neural activity
            and corresponding behavior concatenated.

        behavior: str
            The name of the behavior to use when extracting continuous time
            periods/events.

    Returns:
        epochs: pandas Series
            A Series that has hierarchical indices, where "behavior" is the
            outermost index, followed by "interval". Each interval contains
            an ndarray of timepoints in which the corresponding behavior was
            observed, continuously.
    """
    if behavior not in mouse.spikes_and_beh.columns:
        raise ValueError("'{}' is not a column (i.e. behavior) in the mouse's dataframe".format(behavior))

    dataframe = mouse.spikes_and_beh.copy()

    # Find all timepoints where the behavior discontinues
    dataframe["interval"] = (dataframe[behavior].shift(1) != dataframe[behavior]).astype(int).cumsum()

    # Put the index into the dataframe as a column, without creating a new DataFrame
    dataframe.reset_index(inplace=True)

    # Group the dataframe by behavior and corresponding intervals,
    # and apply np.array to the index column of each group.
    epochs = dataframe.groupby([behavior, "interval"])["index"].apply(np.array)

    return epochs


def filter_epochs(interval_series, framerate=10, seconds=1):
    """Helper function for extract_epochs.

    Args:
        interval_series: pandas Series
            A Series with an ndarray of continous behavior timepoints, for each
            index.

        framerate: int, optional, default: 10
            The framerate that corresponds to the session from which the
            intervals were extracted.

        seconds: int, optional, default: 1
            The amount of seconds.

    Returns:
        intervals: list
            A list of all the intervals that are at least as long as the
            provided framerate multiplied by the provided seconds.
    """
    intervals = [interval for interval in interval_series if len(interval) >= framerate*seconds]
    return intervals

def get_binary_cols(df):
    """
    Helper function that returns the column names of all columns in a dataframe
    that only have boolean values (0 or 1)
    
    Inputs:
        df, pandas dataframe
    
    Returns:
        cols, list of str
            list of boolean column names
    """
    
    cols = []
    for c in df.columns:
        vals = set(df[c])
        if vals==set([0,1]):
            cols.append(c)
    return cols

class Mouse:
    """A base class for keeping all relevant & corresponding objects, i.e.,
    spikes, cell transients, & behavior dataframes, with their respective
    mouse.
    """

    def __init__(self, cell_transients=None, spikes=None, behavior=None, 
                 A=None, Cn=None, **kwargs):
        self.name = kwargs.get("name", None)
        self.age = kwargs.get("age", None)
        self.sex = kwargs.get("sex", None)
        self.FR = kwargs.get('FR', None)
        self.session = kwargs.get('session', None)
        self.group = kwargs.get('group', None)
        self.raw = kwargs.get('raw', None)
        self.cell_transients = cell_transients
        self.spikes = spikes
        self.A = A
        self.Cn = Cn

        if behavior is not None:
            self.behavior = behavior
            self.spikes_and_beh = self.spikes.join(self.behavior, how="left")
        else:
            print("A behavior dataframe was not provided.", file=sys.stderr)

        velocity_cutoff = kwargs.get("velocity_cutoff", None)

        # Adds "Running_frames" column to the end of the behavior Dataframe
        if velocity_cutoff is not None:
            running_frames = np.where(self.behavior["Velocity"] > velocity_cutoff, 1, 0)
            self.behavior = self.behavior.assign(Running_frames=running_frames)
            
    def rates(self,cols=None, in_zone=True):
        """
        Computes the average rate for each neuron during the entire session, and during each
        behavior/zone.
        
        Inputs:
            cols: list, optional
                The column names in Mouse.behavior for which to compute rates.
            in_zone: boolean, optional
                A flag which indicates whether rate calculation should be performed for
                periods during behavior/zone or NOT during behavior/zone. 
        
        """
        
        if cols is None:
#            cols = []
#            for c in self.behavior.columns:
#                vals = set(self.behavior[c])
#                if vals==set([0,1]):
#                    cols.append(c)
            cols = get_binary_cols(self.behavior)
        fr = self.FR
        if fr is None:
            print('No framerate given, defaulting to 10', file=sys.stderr)
            fr = 10
        
        rate_df = pd.DataFrame()
        sb = self.spikes_and_beh
        
        rate_df.loc[:,'entire_session'] = fr * self.spikes.mean()

        #computes the average rate for each neuron during each behavior/zone
        if in_zone == True:
            for c in cols:
                rate_df.loc[:,c] = fr * self.spikes.loc[sb[c] != 0,:].mean()
            rate_df.index.name='neuron'
            rate_df.reset_index(inplace=True)
        #computes the average rate for each neuron when each behavior/zone is NOT occurring
        elif in_zone == False:
            for c in cols:
                rate_df.loc[:,'not_' + c] = fr * self.spikes.loc[sb[c] == 0,:].mean()
            rate_df.index.name='neuron'
            rate_df.reset_index(inplace=True)
        
        if self.name:
            rate_df.loc[:,'mouse'] = np.tile(self.name, len(rate_df)) 
        
        return rate_df

    def rateShuffle(self, resamples=10000, statistic=Resampler.diff_of_mean_rate, cols=None):
        """
        Wrapper method for the Resampler.shuffle function which generates a permutation
        distribution for neuron selectivity analysis.

          Inputs:
            resamples: int, optional
                The number of permutations to perform.
            statistic: Resampler method, optional
                Defines the test statistic to be calculated with each shuffle. Default method
                is a difference of mean rate between in-zone and out-of-zone periods.
            cols: list, optional
                The column names in Mouse.behavior for which to compute rates.
        """
        dataframe = self.spikes
        
        shuffle_dict = {}
        if cols is None:
            cols = get_binary_cols(self.behavior)
        fr = self.FR
        for col in cols:
            shuffle_dict[col] = Resampler.shuffle(resamples, dataframe, statistic, self.behavior[col])
        
        return shuffle_dict
    
    def selectivity(self, p_cutoff=0.05, rate_data_df=None, shuffle_dict=None, cols=None):
        """
        Wrapper method for the cellSelectivity function which calculates single neuron
        selectivity for each neuron for particular behaviors/zones.
        
        Inputs:
            p_value: float, optional
                The p-value cutoff for significant selectivity (i.e. p = 0.05).
                Default is 0.05.
            rate_data_df: pd.DataFrame, optional
                The DataFrame containing single cell event rates for each neuron.
                If none is provided the mouse.rates method will be used.
            shuffle_dict: dict, optional
                A dictionary with key:value pairs of a string indicating the behavior/zone
                and a pd.DataFrame with the permutation distribution for that behavior/zone.
                If none is provided the mouse.rateShuffle method will be used.
            cols: list, optional
                The column names in mouse.behavior for which to compute rates.
        """
        if cols is None:
            cols = get_binary_cols(self.behavior)

        if rate_data_df is None:
            rate_data_df = self.rates()

        if shuffle_dict is None:
            shuffle_dict = self.rateShuffle()

        fr = self.FR

        for col in cols:
            shuffle = shuffle_dict[col]
            classified, z_scores = cellSelectivity(self, shuffle_dict[col], col, p_cutoff=p_cutoff)
            rate_data_df[col + '_selectivity'] = classified
            rate_data_df[col + '_z_score'] = z_scores

        return rate_data_df
    
    def resample(self, new_FR, method='default'):
        """
        Resamples the time series dataframes of a mouse object to a new framerate
        and returns a new mouse object
        
        Inputs:
            new_FR: float
                New frame rate (in seconds)
            method: str, optional
                Method used to resample the dataframes. Must be 'default','mean' or 'sum'
                If 'default', self.spikes will be resampled with 'sum' method, and all others
                will be resampled with 'mean'. This maintains the spike rates and 
                fluorescence values (raw, cell_transients) for plotting purposes.
        Returns:
            resampled_mouse: Mouse object
                New mouse object 
        
        """
        
        import copy
        resampled_mouse = copy.deepcopy(self)
        new_fields = ['spikes','raw','cell_transients','spikes_and_beh','behavior']
    
        for field,data in resampled_mouse.__dict__.items():
            if field in new_fields:
                if method=='default':
                    if field=='spikes':
                        meth = 'sum'
                    else:
                        meth = 'mean'
                else:
                    meth = method
                
                if field in ['spikes_and_beh','behavior']:
                    bin_cols = get_binary_cols(getattr(self,field))
                    # add back Trial_time
                    data_new = resample_ts(data,current_FR=str(np.round(1000/self.FR,3))+'ms',
                                          new_FR=str(np.round(1000/new_FR,3))+'ms',method=meth,drop_time=False)
                    for col in bin_cols:
                        data_new[col] = data_new[col].where(data_new[col] < 0.5, 1)
                        data_new[col] = data_new[col].where(data_new[col] >= 0.5, 0)
                #data = data.fillna(method='ffill')
                else:
                    data_new = resample_ts(data,current_FR=str(np.round(1000/self.FR,3))+'ms',
                          new_FR=str(np.round(1000/new_FR,3))+'ms',method=meth,drop_time=True)
                
                setattr(resampled_mouse,field,data_new)
        resampled_mouse.FR = new_FR
    
        return resampled_mouse

    def reconvolve(self, tau):
        import copy
        new_mouse = copy.deepcopy(self)        
        conv_func = _get_conv_func(tau=tau, FR = self.FR)
        new_mouse.cell_transients = new_mouse.spikes.apply(np.convolve, args=[conv_func,'same'])
        
        return new_mouse    

    def subsample(self,n_neurons,replace=False,random_state=None):
        """
        Returns a Mouse object that contains a subsample of the neurons

        Parameters
        ----------
        n_neurons : int
            The number of neurons to subsample. If n_neurons > the actual number
            of neurons in the Mouse object, the function will return an identical
            Mouse object
        replace : bool, optional
            Sample with replacement. The default is False.
        random_state : int, optional
            The seed that is passed to pandas.DataFrame.sample function.
            The default is None.

        Returns
        -------
        subsampled mouse object.

        """
        import copy
        mouse_samp = copy.deepcopy(self)
        
        if n_neurons < len(self.spikes.columns):
            
            mouse_samp.spikes = self.spikes.sample(n_neurons,axis=1,replace=replace,random_state=random_state)
            mouse_samp.raw = self.raw[mouse_samp.spikes.columns]
            mouse_samp.cell_transients = self.cell_transients[mouse_samp.spikes.columns]
    
        return mouse_samp

def _get_conv_func(tau, FR=10):
    """
    Helper function that returns an exponential decay function to be used
    to re-convolve mouse.cell_transients with the desired tau

    Parameters
    ----------
    tau : float
        decay constant (in seconds) for the convolution function
    FR : int, optional
        Frames per second of the function. The default is 10.

    Returns
    -------
    conv_func : numpy array
        An exponential decay function. It is 1000 frames long, with the
        funtion beginning at frame 500.

    """
    
    n_samples = 500
    tau_frame=tau * FR #tau, in number of frames
    conv_func = np.roll(np.exp(-np.arange(n_samples) / tau_frame), int(n_samples/2))
    conv_func[:int(n_samples/2)] = 0
    
    #conv_func /= np.sum(conv_func)
    
    return conv_func
def cellSelectivity(mouse, permutation_distribution, behavior, statistic=Resampler.diff_of_mean_rate, p_cutoff=0.05):
    """Performs cell selectivity analysis on neural data.

    This function can define cell selectivity for particular behavioral epochs. It computes a test statistic 
    *statistic* (i.e. difference of means) on neural data *mouse.spikes* with a particular behavior/zone *behavior 
    and utilizes a shuffle distribution *permutation_distribution* with a calculated p-value to determine if a cell 
    is 'selective' as defined by a p-value cutoff *p_cutoff (has increased activity in one behavioral epoch vs another 
    behavioral epoch).

    Inputs:
        mouse: Mouse object
            The mouse object from which spike and behavior data are drawn from.
            
        permutation_distribution: Dataframe
            A pandas Dataframe of the permutation distribution.
        
        behavior: string
            The string identifying the behavior/zone used to index the behavior DataFrame.
        
        statistic: function
            The test statistic that is computed on each neuron's activity data. 
            Default is Resampler.diff_of_mean_rate.

        p_cutoff: float
            The p-value cutoff for significant selectivity (i.e. p = 0.05);
            default is 0.05.

    Returns:
        classified: np.array
            An np.array with the selectivity classifications for each neuron during the desired behavior/zone.
        
        z_scores: np.array
            An np.array with the z-scores for each neuron during the desired behavior/zone.
    """
    # draw neural data and framerate from mouse object
    neural_data = mouse.spikes
    fr = mouse.FR
    
    # get the boolean vectors for periods when the mouse is or is not performing a behavior or in a zone
    beh = mouse.behavior[behavior] == 1
    not_beh = mouse.behavior[behavior] == 0
    
    # calculates the test statistic for each neuron
    statistic_vec = statistic(neural_data, beh, not_beh, frame_rate=fr)
    
    # calculates p-values for each neuron
    p_values = np.array([Resampler.p_value(statistic_vec[idx], permutation_distribution[neuron]) for idx,neuron in enumerate(neural_data.columns)])
    
    # calculates z-scores for each neuron
    z_scores = np.array([Resampler.z_score(statistic_vec[idx], permutation_distribution[neuron]) for idx,neuron in enumerate(neural_data.columns)])
    
    #determines the cell selectivity for each neuron
    classified = np.array([None] * len(neural_data.columns))
    classified[(statistic_vec > 0) & (p_values < p_cutoff)] = behavior + '_activated'
    classified[(statistic_vec < 0) & (p_values < p_cutoff)] = behavior + '_inhibited'
    classified[(p_values > p_cutoff)] = 'non_selective'
    
    return classified, z_scores

def downsample_dataframe(dataframe, row_multiple):
    """Downsample a given pandas DataFrame

    Args:
        dataframe: DataFrame
            The pandas DataFrame to be downsampled.

        row_multiple: int
            The row multiple is the rows to be removed,
            e.g., a row_multiple of 3 would remove every 3rd row from the
            provided dataframe

    Returns:
        dataframe: DataFrame
            The downsampled pandas DataFrame.
    """

    # Drop every nth (row multiple) row
    dataframe = dataframe.iloc[0::row_multiple, :]

    # Reset and drop the old indices of the pandas DataFrame
    dataframe.reset_index(inplace=True, drop=True)

    return dataframe

def xy_dist(p1, p2):
    """
    Computes the euclidian distance between two 2D points
    
    Inputs:
        p1,p2: tuple of floats
            The (x,y) location of each point
    Returns:
        dist: float
            The Euclidian distance between p1 and p2
    """    
    
    dist = np.sqrt((p1[0]-p2[0])**2 + (p1[1] - p2[1])**2)
    return dist


def activity_by_neurons(spikes_and_beh, neuron_names, *behaviors, **kwargs):
    """Computes the neuron activity rates for given behaviors

    This function computes the rates for a given animal's activity and
    neuron, given some set of behaviors.

    Args:
        spikes_and_beh: DataFrame
            A concatenated pandas DataFrame of the neuron activity and
            the corresponding behavior, for a given animal.

        neuron_names: list
            The names of the neurons whose rates are to be computed.

        behaviors: arbitrary argument list
            The behaviors for which to compute the activity rates.
            NOTE: If no behaviors are provided, then the average rate over the
            entire session.

        frame_rate: int, optional, default: 10
            The framerate by which to multiply the activity rate.

    Returns:
        activity_df: DataFrame
            A pandas DataFrame of the neuron activity rates.
    """
    if set(neuron_names).issubset(spikes_and_beh.columns) is False:
        raise ValueError("neuron_names is NOT a subset of the columns in spikes_and_beh!")

    frame_rate = kwargs.get("frame_rate", None)

    if frame_rate is None:
        print("You did not specify a frame rate! Defaulting to 10", file=sys.stderr)
        frame_rate = 10

    activity_df = pd.DataFrame(columns=behaviors)

    for behavior in behaviors:
        if behavior in spikes_and_beh.columns:
            activity_df.loc[:, behavior] = frame_rate * spikes_and_beh.loc[spikes_and_beh[behavior] != 0, neuron_names].mean()
        else:
            raise ValueError("{} is not a column (i.e. behavior) in spikes_and_beh.".format(behavior))

    # Return average rate over entire session if no behavior(s) is/are provided
    if not behaviors:
        behavior = "entire_session"
        activity_df = pd.DataFrame(columns=[behavior])
        activity_df.loc[:, behavior] = frame_rate * spikes_and_beh.loc[:, neuron_names].mean()

    return activity_df

def old_div(a, b):
    """
    DEPRECATED: import ``old_div`` from ``past.utils`` instead.

    Equivalent to ``a / b`` on Python 2 without ``from __future__ import
    division``.

    TODO: generalize this to other objects (like arrays etc.)
    
    
    Used in caimain functions
    """

    return a // b


def com(A, d1, d2):
    """Calculation of the center of mass for spatial components

       From Caiman: https://github.com/flatironinstitute/CaImAn
       @author: agiovann

     Inputs:
     ------
     A:   np.ndarray
          matrix of spatial components (d x K)

     d1:  int
          number of pixels in x-direction

     d2:  int
          number of pixels in y-direction

     Output:
     -------
     cm:  np.ndarray
          center of mass for spatial components (K x 2)
    """
    nr = np.shape(A)[-1]
    Coor = dict()
    Coor['x'] = np.kron(np.ones((d2, 1)), np.expand_dims(list(range(d1)),
                        axis=1))
    Coor['y'] = np.kron(np.expand_dims(list(range(d2)), axis=1),
                        np.ones((d1, 1)))
    cm = np.zeros((nr, 2))        # vector for center of mass
    cm[:, 0] = old_div(np.dot(Coor['x'].T, A), A.sum(axis=0))
    cm[:, 1] = old_div(np.dot(Coor['y'].T, A), A.sum(axis=0))

    return cm

def behavior_start_stop(beh_intervals, time, fr):
        
    beh_starts = [beh_intervals[i][0] for i in range(len(beh_intervals))]
    beh_stops = [beh_intervals[i][-1] for i in range(len(beh_intervals))]
    start_range = [[i-time*fr,i+time*fr] for i in beh_starts]
    stop_range = [[i-time*fr,i+time*fr] for i in beh_stops]

    return start_range, stop_range

def event_triggered_activity(mice, behavior=None, time=5, min_time=1):
    beh_intervals = {}

    activity_start = {} #dict storing event-start activity
    activity_stop = {} #dict storing event-stop activity

    for k,v in mice.items():
        if behavior in mice[k].behavior.columns:

            beh_intervals[k] = extract_epochs(mice[k],behavior)[1]
            beh_intervals[k] = filter_epochs(beh_intervals[k],framerate=mice[k].FR,seconds=min_time)

            #get time ranges for behavior start and stop 
            start_range, stop_range = behavior_start_stop(beh_intervals[k], time, v.FR)

            #define the time points which to analyze
            xax = np.arange(-time,time,1/v.FR)

            #initialize activity arrays
            act_start_array = np.zeros(shape=(len(xax), len(mice[k].cell_transients.columns), 1))
            act_stop_array = np.zeros(shape=(len(xax), len(mice[k].cell_transients.columns), 1))
            
            #get all neural activity during the start and stop time windows
            for i in start_range:
                act_start = np.array(v.cell_transients.loc[i[0]:i[1]-1,:].reset_index(drop=True))[:,:,np.newaxis]
                if (len(act_start) == len(xax)):
                    act_start_array = np.concatenate([act_start_array, act_start], axis=2)

            for i in stop_range:
                act_stop = np.array(v.cell_transients.loc[i[0]:i[1]-1,:].reset_index(drop=True))[:,:,np.newaxis]
                if (len(act_stop) == len(xax)):
                    act_stop_array = np.concatenate([act_stop_array, act_stop], axis=2)

            activity_start[k] = act_start_array
            activity_stop[k] = act_stop_array
            
    return activity_start, activity_stop, xax

def get_eta(activity, xax, which_cells):
    
    temp = []
    for m in which_cells['mouse'].unique():
        behavior = which_cells['Behavior'].unique()[0]
        neurons = which_cells.query(f"mouse == '{m}'")['neuron'].values - 1
        
        temp.append(activity[m, behavior][:,neurons,:].mean(axis=2))
    
    temp = tuple(temp)
    
    eta = np.concatenate(temp, axis=1)
        
    return eta   

def plot_eta(eta, xax, ax, color='k'):
    mu = np.mean(eta,1)
    se = np.std(eta,1)/np.sqrt(eta.shape[1])
    #fig,ax = plt.subplots(figsize=(2.5,2.5))

    ax.plot(xax,mu, linewidth=1, color=color)
    ax.fill_between(xax,mu+se,mu-se, zorder=0, alpha=.3, color=color)
    ax.axvline(x=0,color='r',linestyle='--')
    ax.set_xlabel('Time Relative \nto Behavior (s)')
    ax.set_ylabel('Average Z-scored \nActivity')

def analyze_eta(mice, behavior, ax=None, time=5, min_time=1, which_cells=[], colors=['k','b']):
    
    if ax is None:
        fig,ax = plt.subplots(1, 2, figsize=(4,2))
    colors = colors
    
    activity_start, activity_stop, xax = event_triggered_activity(mice=mice, behavior=behavior, time=time, min_time=min_time)
    
    for i, cells in enumerate(which_cells):
        eta_start = get_eta(activity_start, xax, cells)
        eta_stop = get_eta(activity_stop, xax, cells)


        plot_eta(eta_start, xax, ax[0], color=colors[i])
        ax[0].set_title(behavior+' start')
        
        plot_eta(eta_stop, xax, ax[1], color=colors[i])
        ax[1].set_title(behavior+' stop')
        ax[1].set_ylabel('')
    return ax
    
def resample_ts(timeseries, current_FR, new_FR, method='mean',drop_time=True):
    timeseries = timeseries.assign(Trial_time=pd.timedelta_range(0, periods=len(timeseries.index), freq=current_FR))
    if method=='mean':
        timeseries = timeseries.resample(new_FR, on="Trial_time").mean().reset_index(drop = drop_time)
    elif method=='sum':
        timeseries = timeseries.resample(new_FR, on="Trial_time").sum().reset_index(drop = drop_time)
    else:
        print('Unsopported resampling method: please choose mean or sum')
        return;
    return timeseries

# Analysis Functions for Single Neuron Properties

def selectivity_counts(neuron_data_df, behaviors, normalize=False):
    """Determines the frequency of activated, inhibited, and non-selective neurons for specified behaviors/zones.
    
    This function takes a DataFrame containing information regarding single neurons (e.g. selectivity_data_df)
    and calculates the frequency of activated, inhibited, and non-selective neurons for each specified
    behavior/zone.

    Inputs:
        neuron_data_df: pd.DataFrame
            A pandas DataFrame containing information (e.g. rates, selectivity, z-scores, other metrics) regarding
            single neurons.
        behaviors: list
            A list of strings which define the behaviors or zones of interest. These should correspond to column
            names in neuron_data_df
        normalize: boolean
            Flag which indicates whether absolute or relative frequencies should be calculated.
            Default is False which produces absolute frequencies.
        
    Returns:
        selectivity_counts_df: pd.DataFrame
            A 'wide' pandas DataFrame with the specified behaviors as columns and 
            index = ['activated', 'inihibited', 'non_selective']. Values are either the absolute or 
            relative frequencies.
    """
    selectivity_col_names = [behavior + '_selectivity' for behavior in behaviors]
    selectivity_counts_df = neuron_data_df[selectivity_col_names].replace(
        {r'(^.*activated.*$)':'activated', r'(^.*inhibited.*$)':'inhibited'}, regex=True).apply(
        pd.value_counts, normalize=normalize).fillna(0)
    selectivity_counts_df.columns = [col_name.replace('_selectivity', '') for col_name in selectivity_counts_df.columns]
    
    if normalize == False:
        selectivity_counts_df = selectivity_counts_df.astype('int')
    
    return selectivity_counts_df
    
    
    #dfs = [] # initialize an empty list to accumulate processed counts DataFrames for each behavior/zone
    #for behavior in behaviors:
        # change the column labels to remove the string '_selectivity' and the indices to remove behavior/zone
    #    temp = neuron_data_df[behavior+'_selectivity'].value_counts(normalize=normalize).rename(
    #        index={behavior+'_activated':'activated', behavior+'_inhibited':'inhibited'}).sort_index()
    #    temp.name = behavior
    #    dfs.append(temp)
    #selectivity_counts_df = pd.concat(dfs, axis=1, sort=True).fillna(0)
    
def PairwiseCategoricalOverlap(categorical_df, statistical_test=fisher_exact):
    # Generate the index by using the cartesian product of each column name with the unique entries in the column
    ind_list = []
    for name, values in categorical_df.items():
        ind_list.extend(list(product([name], np.sort(values.unique()))))
    overlap_index = pd.MultiIndex.from_tuples(ind_list) # Create a pandas MultiIndex
    
    # Create empty DataFrames to collect p-values and odds ratios using the index generated above as both the index 
    # and column labels
    p_value_df = pd.DataFrame(index=overlap_index, columns=overlap_index)
    odds_ratio_df = pd.DataFrame(index=overlap_index, columns=overlap_index)
    
    # Generate a 2x2 table for each pair of categorical values and compute the p-value and odds ratio using a
    # statistical test which should take as input the 2x2 table and output an odds ratio and p-value
    for ind in overlap_index:
        for col in overlap_index:
            tab = pd.crosstab((categorical_df[ind[0]] == ind[1]).astype('str'), 
                              (categorical_df[col[0]] == col[1]).astype('str')).loc[['True', 'False'],['True','False']]
            odds_ratio_df.loc[ind,col], p_value_df.loc[ind,col] = statistical_test(tab)
    
    # Typecast the p-value and odds ratio DataFrames to floats
    p_value_df = p_value_df.astype('float')
    odds_ratio_df = odds_ratio_df.astype('float')
    
    return p_value_df, odds_ratio_df

def SelectivityOverlap(neuron_data_df, behaviors, statistical_test=fisher_exact):
    
    selectivity_cat_df = neuron_data_df[[behavior+'_selectivity' for behavior in behaviors]].copy()
    selectivity_cat_df.replace({r'(^.*activated.*$)':'activated', r'(^.*inhibited.*$)':'inhibited'}, regex=True, inplace=True)
    selectivity_cat_df.columns = [colname.replace('_selectivity', '') for colname in selectivity_cat_df.columns]
    
    p_value_df, odds_ratio_df = PairwiseCategoricalOverlap(selectivity_cat_df, statistical_test=statistical_test)
    
    return p_value_df, odds_ratio_df
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  5 15:15:17 2018

Ca transient event detection, adapted from Dombeck 2007

This is based on the MATLAB implementation of detect_ca_transients_mossy which
was written by Jessica Jimenez, Columbia University, jcj2123@columbia.edu

@author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
"""
import math
import numpy as np
import pandas as pd

from analysis.analysis_utils import z_score_data

class Deconvoluter:
    """
    Attributes:

        cell_transients: DataFrame
            T x N matrix with s.d. vlaues for all the timepoints of the
            qualified transients (all zeros except for transients)

        cell_data: DataFrame
            T x N matrix with all z-scored data.

        cell_auc_df: DataFrame
            T x N matrix with computed transient area under the curve (AUC)
            values (all zeros except for the AUC value assigned to the peak
            timepoint of each transient).
    """

    def __init__(self, raw_data, **kwargs):
        """
        Args:
            raw_data: DataFrame
                Ca2 transient data in T x N format, where N is the # of neuron
                column vectors, and T is the number of observations (rows),
                all in raw format.

            threshold: int, optional, default: 2
                The minimum aplitude size of ca transient data in s.d.

            baseline: float, optional, default: 0.5
                The standard deviation offset value of Ca transient.

            t_half: float, optional, default: 0.2
                The half life of gcamp type used (s).

            frame_rate: int, optional, default: 10
                The frame rate for the data.
        """

        threshold = kwargs.get("threshold", 2)
        baseline = kwargs.get("baseline", 0.5)
        t_half = kwargs.get("t_half", 0.2)
        frame_rate = kwargs.get("frame_rate", 10)

        # z-score all raw data
        self.cell_data = z_score_data(raw_data)

        # Preallocate dataframes
        self.cell_transients = pd.DataFrame(np.zeros((len(raw_data.index), len(raw_data.columns))))
        self.cell_auc_df = pd.DataFrame(np.zeros((len(raw_data.index), len(raw_data.columns))))

        # Define minimum duration of calcium transient based on gcamp type used

        # Simplified from (-ln(A / Ao) / t_half), [A / Ao] = 0.5 at t half-life,
        # [-ln(A / Ao)] = 0.693
        decay_rate = np.log(2) / t_half

        # Minimum (s) duration for ca transient of minimum
        # specified s.d. amplitude threshold
        minimum_duration = -(math.log(baseline / threshold)) / decay_rate

        # Minimum number of frames the ca transient should last above baseline
        self.minimum_frames = round(minimum_duration * frame_rate)

        self.detect_ca_transients_mossy(threshold, baseline)

    def detect_ca_transients_mossy(self, threshold, baseline):
        """Preprocesses neural signal data obtained via Calcium Imaging

        Args:
            threshold: int
                The minimum aplitude size of ca transient data in s.d.

            baseline: float
                The standard deviation offset value of Ca transient.
        """

        # Identify qualified ca transients and generate outputs
        for column in self.cell_data:

            # Find all timepoints where flourescence greater than threshold
            onset = self.cell_data.loc[self.cell_data[column] > threshold].index

            # Find all timepoints where floursecence greater than the
            # baseline (transient offset)
            offset = self.cell_data.loc[self.cell_data[column] > baseline].index

            found = True

            for i in range(0, len(offset)-1):
                if found:

                    # Specify start index of transient from offset vector
                    start = offset[i]

                    found = False

                # Specify stop index of transient from offset vector
                if offset[i+1] - offset[i] > 1 or i == len(offset)-2:

                    # Deals with the last index in the vector
                    if i == len(offset)-2:
                        finish = offset[i+1]
                    else:
                        finish = offset[i]

                    # Find the peak value in that start-stop range
                    # peak_value = cell_data.loc[start:finish+1, column].max()
                    peak_index = self.cell_data.loc[start:finish+1, column].values.argmax()
                    transient_vector = np.arange(start, finish+1)

                    # Retrieve "cell" index of that peak value
                    max_amp_index = transient_vector[peak_index]

                    # peak_to_offset_vector = list(range(max_amp_index, finish+1))
                    found = True

                    # If the peak value index from start-stop in offset is also
                    # found in onset vector, then the transient exceeded the
                    # 2SD threshold
                    if ((finish+1) - max_amp_index) >= self.minimum_frames and (max_amp_index in onset):

                        # Retrieve "cell" values for all the timepoints of
                        # that transient.
                        self.cell_transients.loc[start:finish+1, column] = self.cell_data.loc[start:finish+1, column]

                        # Create matrix with all calcium transient peak events
                        # (all 0's except for amplitude value at peak timepoint)
                        # cell_events(maxamp_ind, k) = peak_value

                        # Integrate area under curve of transient from start-stop
                        transient_area = np.trapz(self.cell_data.loc[start:finish+1, column])

                        # Add all the calcium transient Area Under Curve values
                        # (all 0's except for AUC value assigned to peak
                        # timepoint)
                        self.cell_auc_df.loc[max_amp_index, column] = transient_area



"""
Miscellaneous, or unfinished functions written by Jack Berry 
to be used for calcium imaging analysis
3/1/2019
"""
import numpy as np
import pandas as pd
import analysis.analysis_utils as au

##### Behavior Functions #######


def read_ethovision(etho_filepath, track_point):
    """
    Reads in data from an Ethovision raw data file (converted to .csv format). Expected format is a T x B matrix,
    where T is time bins, and B is the automatically tracked behaviors.
    
    Intended to to be used as a helper function for preprocess_behavior
    
    Args:
        etho_filepath: str
            The file path to the raw Ethovision .csv file
            
        track_point: str
            The tracking point to be used (options: 'center', 'nose', 'tail')
            
    Returns:
        behavior: pandas DataFrame
            The processed Ethovision data    
    """       
    
    # read ethovision data and rename columns
    nrowsheader = pd.read_csv(etho_filepath, nrows=0, encoding="latin1", header=0)
    nrowsheader = int(nrowsheader.columns[1])

    behavior = pd.read_csv(etho_filepath, header=nrowsheader-2, encoding="latin1", low_memory=False)
    behavior = behavior.drop(behavior.index[0])
    behavior = behavior.apply(pd.to_numeric, errors='coerce')

    # remove columns that track other body points (ie. if track_point = 'center', remove all nose and tail columns)
    all_points = set(['center', 'nose', 'tail'])
    unwanted_points = all_points - set([track_point])
    cols = [c for c in behavior.columns if not any(s in c for s in unwanted_points)]
    behavior = behavior[cols]       
    
    #each track_point has a different string in the column name that should be removed
    track_point_dict = {"center": "_/_center-point)", "nose": "_/_nose-point)", "tail": "_/_tail-base" }
    track_point_dict_etho14 = {"center": "(center-point)", "nose": "(nose-point)","tail": "(tail-base)"}
    # rename columns
    new_cols = []
    for col in behavior.columns:
        col = col.replace(' ', '_').replace("In_zone(", '').replace(track_point_dict[track_point], '')
        col = col.replace(track_point_dict_etho14[track_point], '')
        new_cols.append(col)
    behavior.columns = new_cols

    old_x = 'X_' + track_point
    old_y = 'Y_' + track_point
    #for now, keep _center so spatial info code will work
    behavior = behavior.rename(columns = {old_x : 'X_center', old_y : 'Y_center'})
    
    # convert Trial_time to timedelta function to enable time operations    
    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')

    return behavior

def read_observer(observer_filepath):
    """
    Reads in data from an observer file (converted to .csv format). Expected format is a T x B matrix,
    where T is time bins, and B is the hand-scored behaviors.
    
    Intended to to be used as a helper function for preprocess_behavior
    
    Args:
        observer_filepath: str
            The file path to the raw observer .csv file
            
    Returns:
        behavior: pandas DataFrame
            The processed observer data
    
    """
    behavior = pd.read_csv(observer_filepath)
    behavior = behavior.drop(["Observation", "Event Log", "Time"], axis=1)
    behavior.index += 1
    behavior.fillna(0, inplace=True)
    behavior = behavior.apply(pd.to_numeric, errors='coerce')
    behavior[behavior != 0] = 1
    
    return behavior

def make_freq_string(fr):
    # converts the framerate to a frequency in ms
    freq = np.round(1 / fr * 1000, 3)
    freq_string = str(freq) + 'ms'
    
    return freq_string

def read_boris(boris_filepath, freq_string):
    # read in from csv (BORIS or stopwatch) the start and stop times for various behaviors
    csv = pd.read_csv(boris_filepath, header = [15])
    rec_length = csv['Total length'][0]
    frames = csv['FPS'][0] * (rec_length)
    frames = frames.astype('int')
    
    # make a timedelta array that is as long as the recording time of the behavior video, use 30fps to generate the array
    # assumes frame rate of 30fps here
    time = pd.timedelta_range(0, periods=frames, freq=freq_string)
    #make dt column
    dt = pd.DataFrame(np.roll(time, -1) - time).shift(1)

    
    # make a dataframe with timedelta array as the index
    beh = pd.DataFrame(np.zeros(shape=(len(time),len(csv['Behavior'].unique())),dtype='int'))
    beh = beh.set_index(time)
    beh.index.names = ['Trial_time']
    beh.columns = csv['Behavior'].unique()
    
    # convert start stop times to timedelta
    pairs_raw = pd.to_timedelta(csv['Time'], unit = 's')
    
    # convert timedeltas from the csv to the closest timedelta in the full length timedelta array 
    for i in pairs_raw:
        pairs_raw[pairs_raw == i] = time[np.abs(time-i).argmin()]
        
    # make pairs_raw timedelta the index for the behavior and start-stop arrays
    data = csv[['Behavior', 'Unnamed: 8']]
    data.index = pairs_raw
    
    # generate pairs of indices to slice the full length timedelta array
    for b in data['Behavior'].unique():
        temp = data.query("Behavior == @b")
        pairs = list(zip(temp.index[::2], temp.index[1::2]))
        for pair in pairs:
            beh.loc[pair[0]:pair[1], b] = 1
        
    return beh

def preprocess_behavior(etho_filepath=None, observer_filepath=None, track_point='center'):
    """
    Processes original ethovision and observer files (converted to .csv format)
    Args:
        etho_filepath: str
            The file path to raw ethovision behavior .csv file.
        observer_filepath: str
            The file path to the raw observer .csv file.
    Returns:
        behavior: DataFrame
            The preprocessed behavior DataFrame.
    """

    if etho_filepath and observer_filepath:
        #Read in behavior files
        etho_behavior = read_ethovision(etho_filepath, track_point)
        obs_behavior = read_observer(observer_filepath)
        num_obs = len(obs_behavior.columns)
        
        #Merge Ethovision and Observer data
        behavior = pd.merge(etho_behavior, obs_behavior, how="left", left_index=True, right_index=True)
        behavior.update(behavior[behavior.columns[-num_obs:]].fillna(0))
        
        return behavior

    if etho_filepath:
        etho_behavior = read_ethovision(etho_filepath, track_point)
        return etho_behavior
        
    if observer_filepath:
        obs_behavior = read_observer(observer_filepath)
        return obs_behavior
    
    raise ValueError("You did not provide an ethovision or observer file path")


def process_EPM_behavior(behavior, analysis_frame_length = '100ms', hd_open=False):
    """
    Process behavior dataframe containing tracking information from an EPM experiment.
    Assumes there is not already a Center zone, and assumes the mouse is in the 
    center whenever it is not in closed or open arms. 
    
    Inputs:
        behavior: T x B Pandas DataFrame
            - raw Ethovision (with or without manual scoring from Observer)
            behavior tracking file for EPM.
        analysis_frame_length: str
            - desired frame length of the final behavior dataframe. eg. "100ms"
        hd_open: bool, optional
            - if True, all time bins during which the mouse is headdiping will
            be considered OpenArms. This fix can be used to correct times when
            the mouse is tracked to be in the center or closed arms, but is
            in actually investigating the open arms.
            - if False, normal tracking will be used
    Returns:
        behavior: T x B Pandas DataFrame
            - processed behavior dataframe that has been downsampled, interpolated
            and new columns added
    """

    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')
    behavior.set_index('Trial_time', inplace=True, drop=True)
    behavior = behavior.resample(analysis_frame_length).mean()
    #behavior = behavior.resample("100ms", on="Trial_time").mean()
    
    #After resampling, zone columns no longer contain only 0 and 1 (ie. when 
    #mouse transitioned between frames. To avoid ambiguity, round values to 0 or 1
    
    varlist = ['OpenArms', 'ClosedArms', 'Open1', 'Open2', 'Closed1', 'Closed2', 'Headdips']
    for var in varlist:
        if var in behavior.columns:
            behavior[var] = behavior[var].fillna(method='bfill')
            behavior[var] = behavior[var].where(behavior[var] < 0.5, 1)
            behavior[var] = behavior[var].where(behavior[var] >= 0.5, 0)
    
    #remove NaNs in X,Y due to point being outside of Arena (Ethovision doesn't interpolate in this situation) by interpolating  

    behavior.X_center = behavior.X_center.interpolate()
    behavior.Y_center = behavior.Y_center.interpolate()
    
    x=behavior.X_center.astype(float)
    y=behavior.Y_center.astype(float)
    behavior['Distance_moved'] = au.distance_moved(x,y)
    behavior['Velocity'] = au.compute_velocity(behavior.Distance_moved)    
    
    #Add column to identify immobile time bins (min_vel = 2, min_dur=1)

    behavior['immobile'] = au.define_immobility(behavior.Velocity)

    #Add Center and other columns to behavior df

    center = (behavior["OpenArms"] + behavior["ClosedArms"])
    center = 1-center
    behavior.loc[:,'Center'] = center
    behavior.loc[:,"Open_Center"] = behavior["OpenArms"] + behavior["Center"]
    
    if 'Headdips' in behavior.columns:
        #Add not Headdips, Center_Headdips, Open+Headdips
        behavior.loc[behavior['Headdips'] == 0, 'no_Headdips'] = 1
        behavior['no_Headdips'] = behavior['no_Headdips'].fillna(0).astype(int)
        behavior.loc[behavior['Headdips'] +  behavior['Center']==2,'Center_Headdips'] = 1
        behavior['Center_Headdips'] = behavior['Center_Headdips'].fillna(0)
        
        if hd_open:
            
            behavior.loc[behavior['Center_Headdips']==1,'Center'] = 0
            behavior.loc[behavior['Center_Headdips']==1,'OpenArms'] = 1
    
    
    behavior.reset_index(inplace = True)
    
    return behavior

def process_ABAB_behavior(behavior):
    """
    Process behavior dataframe containing tracking information from an EPM experiment.
    Assumes there is not already a Center zone, and assumes the mouse is in the 
    center whenever it is not in closed or open arms. 
    
    Inputs:
        behavior: T x B Pandas DataFrame
            - raw Ethovision (with or without manual scoring from Observer)
            behavior tracking file for EPM.
            
    Returns:
        behavior: T x B Pandas DataFrame
            - processed behavior dataframe that has been downsampled, interpolated
            and new columns added
    """

    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')
    behavior.set_index('Trial_time', inplace=True, drop=True)
    behavior = behavior.resample('100ms').mean()
    #behavior = behavior.resample("100ms", on="Trial_time").mean()
    
    #After resampling, zone columns no longer contain only 0 and 1 (ie. when 
    #mouse transitioned between frames. To avoid ambiguity, round values to 0 or 1
    
    varlist = []
    for var in varlist:
        if var in behavior.columns:
            behavior[var] = behavior[var].where(behavior[var] < 0.5, 1)
            behavior[var] = behavior[var].where(behavior[var] >= 0.5, 0)
    
    #remove NaNs in X,Y due to point being outside of Arena (Ethovision doesn't interpolate in this situation) by interpolating  

    behavior.X_center = behavior.X_center.interpolate()
    behavior.Y_center = behavior.Y_center.interpolate()
    
    x=behavior.X_center.astype(float)
    y=behavior.Y_center.astype(float)
    behavior['Distance_moved'] = au.distance_moved(x,y)
    behavior['Velocity'] = au.compute_velocity(behavior.Distance_moved)    
    
    #Add column to identify immobile time bins (min_vel = 2, min_dur=1)

    behavior['immobile'] = au.define_immobility(behavior.Velocity)
    
    behavior.reset_index(inplace = True)
    
    return behavior

def process_OFT_behavior(behavior, analysis_frame_length = '100ms'):
    """
    Process behavior dataframe containing tracking information from an OFT experiment.
    Assumes there is a Center and Periphery zone. 
    
    Inputs:
        behavior: T x B Pandas DataFrame
            - raw Ethovision (with or without manual scoring from Observer)
            behavior tracking file for OFT.
        analysis_frame_length: str
            - desired frame length of the final behavior dataframe. eg. "100ms"
    Returns:
        behavior: T x B Pandas DataFrame
            - processed behavior dataframe that has been downsampled, interpolated
            and new columns added
    """

    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')
    behavior.set_index('Trial_time', inplace=True, drop=True)
    behavior = behavior.resample(analysis_frame_length).mean()
    #behavior = behavior.resample("100ms", on="Trial_time").mean()
    
    #After resampling, zone columns no longer contain only 0 and 1 (ie. when 
    #mouse transitioned between frames. To avoid ambiguity, round values to 0 or 1
    
    varlist = ['center','periphery']
    for var in varlist:
        if var in behavior.columns:
            behavior[var] = behavior[var].where(behavior[var] < 0.5, 1)
            behavior[var] = behavior[var].where(behavior[var] >= 0.5, 0)
    
    #remove NaNs in X,Y due to point being outside of Arena (Ethovision doesn't interpolate in this situation) by interpolating  

    behavior.X_center = behavior.X_center.interpolate()
    behavior.Y_center = behavior.Y_center.interpolate()
    
    x=behavior.X_center.astype(float)
    y=behavior.Y_center.astype(float)
    behavior['Distance_moved'] = au.distance_moved(x,y)
    behavior['Velocity'] = au.compute_velocity(behavior.Distance_moved)    
    
    #Add column to identify immobile time bins (min_vel = 2, min_dur=1)

    behavior['immobile'] = au.define_immobility(behavior.Velocity)

    behavior.reset_index(inplace = True)
    
    return behavior

########## Below are functions to import AnyMaze files

def read_AnyMaze(AM_filepath):
    """
    Reads in data from an AnyMaze raw data file (in to .csv format). Expected format is a T x B matrix,
    where T is time bins, and B is the automatically tracked behaviors (i.e., Open Arm 1, etc.).
    
    Intended to to be used as a helper function for preprocess_behavior_AM
    
    Args:
        AM_filepath: str
            The file path to the raw AnyMaze .csv file
            
    Returns:
        behavior: pandas DataFrame
            The processed AnyMaze data    
    """   
    
    # read AnyMaze data 
    nrowsheader = pd.read_csv(AM_filepath, nrows=0, encoding="latin1", header=0)
    # nrowsheader = int(nrowsheader.columns[1])
    behavior = pd.read_csv(AM_filepath)
    
    # drop nan's and 1st row 
    behavior = behavior.drop(behavior.index[0]) 
    #behavior = behavior.apply(pd.to_numeric, errors='coerce')
    behavior = behavior.replace([np.inf, -np.inf], np.nan) # drop inf and NaN's
    behavior = behavior.dropna()  

    # rename columns of anymaze file
    renamed = []
    for i in behavior.columns:
        i = i.replace(' ','_').replace('Centre_position_','').replace('In_','').replace('_arm_','').replace('_Arm_','')
        renamed.append(i)
    behavior.columns = renamed  
    
    behavior = behavior.rename(columns={'Time':'Trial_time', 'X':'X_center','Y':'Y_center'})
    
    # add up the 2 separate files for closed and open arms
    behavior['Closed_Arms'] = behavior['Closed1'] + behavior['Closed2']
    behavior['Open_Arms'] = behavior['Open1'] + behavior['Open2']
    
    # convert Trial_time to timedelta function to enable time operations
    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')

    return behavior

def preprocess_behavior_AM(AM_filepath=None):
    """
    Processes original AnyMaze files (converted to .csv format)

    Args:
        AM_filepath: str
            The file path to raw AnyMaze behavior .csv file.

    Returns:
        behavior: DataFrame
            The preprocessed behavior DataFrame.
    """

    if AM_filepath:
        #Read in behavior files
        AM_behavior = read_AnyMaze(AM_filepath)
        
        behavior = AM_behavior

        return behavior
    
    raise ValueError("You did not provide an AnyMaze file path.")
    
    
def process_EPM_behavior_AM(behavior, analysis_frame_length = '200ms'):
    """
    Process behavior dataframe containing tracking information from an EPM experiment.
    
    Inputs:
        behavior: T x B Pandas DataFrame
            - raw AnyMaze file (with or without manual scoring from Observer)
            behavior tracking file for EPM.
            **NOTE that there are no 'headdip' analyses for AnyMaze files 
        analysis_frame_length: str
            - desired frame length of the final behavior dataframe. eg. "200ms"
        hd_open: bool, optional
            - if True, all time bins during which the mouse is headdiping will
            be considered OpenArms. This fix can be used to correct times when
            the mouse is tracked to be in the center or closed arms, but is
            in actually investigating the open arms.
            - if False, normal tracking will be used
    Returns:
        behavior: T x B Pandas DataFrame
            - processed behavior dataframe that has been downsampled, interpolated
            and new columns added
    """
    
    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')
    behavior.set_index('Trial_time', inplace=True, drop=True) 
    behavior = behavior.resample(analysis_frame_length).mean()
    #behavior = behavior.resample(analysis_frame_length, on='Trial_time').mean() # this usually messes up the fxn
    
    #After resampling, zone columns no longer contain only 0 and 1 (ie. when 
    #mouse transitioned between frames. To avoid ambiguity, round values to 0 or 1
    
    varlist = ['Closed1', 'Open1', 'Center', 'Closed2', 'Open2', 'Closed_Arms', 'Open_Arms']
    for var in varlist:
        if var in behavior.columns:
            behavior[var] = behavior[var].where(behavior[var] < 0.5, 1)
            behavior[var] = behavior[var].where(behavior[var] >= 0.5, 0)

    #remove NaNs in X,Y due to point being outside of Arena (AnyMaze MAY ALREADY) by interpolating  
    #behavior.X_center = behavior.X_center.interpolate()
    #behavior.Y_center = behavior.Y_center.interpolate()
    
    x=behavior.X_center.astype(float)
    y=behavior.Y_center.astype(float)
    behavior['Distance_moved'] = au.distance_moved(x,y)
    behavior['Velocity'] = au.compute_velocity(behavior.Distance_moved)   
    
    behavior=behavior.dropna()
    
    #Add column to identify immobile time bins (min_vel = 2, min_dur=1)
    behavior['immobile'] = au.define_immobility(behavior.Velocity)
    
    behavior=behavior.reset_index(inplace = False)
    
    return behavior

##### Below are functions to import StopWatch files #####

def read_StopWatch(SW_filepath):
    """
    Reads in data from an StopWatch raw data file (in to .csv format). Expected format is a T x B matrix,
    where T is time bins, and B is the automatically tracked behaviors (i.e., freezing).
    
    NOTE: The StopWatch file (original) should have been converted previously in Matlab. 
    
    Intended to to be used as a helper function for preprocess_behavior_SW
    
    Args:
        SW_filepath: str
            The file path to the converted StopWatch .csv file
            
    Returns:
        behavior: pandas DataFrame
            The processed StopWatch data    
    """   
    
    # read StopWatch data 
    nrowsheader = pd.read_csv(SW_filepath, nrows=0, encoding="latin1", header=0)
    # nrowsheader = int(nrowsheader.columns[1])
    behavior = pd.read_csv(SW_filepath)
    
    # drop nan's and 1st row 
    behavior = behavior.drop(behavior.index[0]) 
    #behavior = behavior.apply(pd.to_numeric, errors='coerce')
    behavior = behavior.replace([np.inf, -np.inf], np.nan) # drop inf and NaN's
    behavior = behavior.dropna()  
    
    # rename the time trial only
    behavior = behavior.rename(columns={'time':'Trial_time'})
    
    
    # convert Trial_time to timedelta function to enable time operations
    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')

    return behavior

def preprocess_behavior_SW(SW_filepath=None):
    """
    Processes original StopWatch files (converted to .csv format)

    Args:
        SW_filepath: str
            The file path to raw StopWatch behavior .csv file.

    Returns:
        behavior: DataFrame
            The preprocessed behavior DataFrame.
    """

    if SW_filepath:
        #Read in behavior files
        SW_behavior = read_StopWatch(SW_filepath)
        
        behavior = SW_behavior

        return behavior
    
    raise ValueError("You did not provide a StopWatch file path.")
    
    
def process_CFC_behavior_SW(behavior, analysis_frame_length = '200ms'):
    """
    Process behavior dataframe containing tracking information from a CFC experiment.
    
    Inputs:
        behavior: T x B Pandas DataFrame
            - StopWatch file, converted from its original to an "AnyMaze" like format
            on Matlab
        analysis_frame_length: str
            - desired frame length of the final behavior dataframe. eg. "200ms"
    Returns:
        behavior: T x B Pandas DataFrame
            - processed behavior dataframe that has been downsampled, interpolated
            and new columns added
    """
    
    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')
    behavior.set_index('Trial_time', inplace=True, drop=True) 
    behavior = behavior.resample(analysis_frame_length).mean()
    #behavior = behavior.resample(analysis_frame_length, on='Trial_time').mean() # this usually messes up the fxn
    
    #After resampling, zone columns no longer contain only 0 and 1 (ie. when 
    #mouse transitioned between frames. To avoid ambiguity, round values to 0 or 1
    
    varlist = ['freezing']
    for var in varlist:
        if var in behavior.columns:
            behavior[var] = behavior[var].where(behavior[var] < 0.5, 1)
            behavior[var] = behavior[var].where(behavior[var] >= 0.5, 0)

    # add a column to the behavioral dataframe that indicates "moving" 
    behavior['moving']=~behavior['freezing']
    
    behavior=behavior.dropna()
    
    behavior=behavior.reset_index(inplace = False)
    
    return behavior

def process_TST_behavior_SW(behavior, analysis_frame_length = '200ms'):
    """
    Process behavior dataframe containing tracking information from a TST experiment.
    The files should be in a StopWatch (hand-scored) format.
    
    Inputs:
        behavior: T x B Pandas DataFrame
            - StopWatch file, converted from its original to an "AnyMaze" like format
            on Matlab
        analysis_frame_length: str
            - desired frame length of the final behavior dataframe. eg. "200ms"
    Returns:
        behavior: T x B Pandas DataFrame
            - processed behavior dataframe that has been downsampled, interpolated
            and new columns added
    """
    
    behavior.Trial_time = pd.to_timedelta(behavior.Trial_time, unit = 's')
    behavior.set_index('Trial_time', inplace=True, drop=True) 
    behavior = behavior.resample(analysis_frame_length).mean()
    #behavior = behavior.resample(analysis_frame_length, on='Trial_time').mean() # this usually messes up the fxn
    
    #After resampling, zone columns no longer contain only 0 and 1 (ie. when 
    #mouse transitioned between frames. To avoid ambiguity, round values to 0 or 1
    
    varlist = ['freezing']
    for var in varlist:
        if var in behavior.columns:
            behavior[var] = behavior[var].where(behavior[var] < 0.5, 1)
            behavior[var] = behavior[var].where(behavior[var] >= 0.5, 0)

    # add a column to the behavioral dataframe that indicates "moving" 
    behavior['moving']=~behavior['freezing']
    
    behavior=behavior.dropna()
    
    behavior=behavior.reset_index(inplace = False)
    
    return behavior



"""
A simple example for VLC python bindings using PyQt5.

Author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
Date: 25 December 2018
"""

import platform
import queue
import os
import subprocess
import sys

from PyQt5 import QtWidgets, QtGui, QtCore
import vlc
from server import Server
from neuron_overlay import Overlay


class Player(QtWidgets.QMainWindow):
    """A simple Media Player using VLC and Qt
    """

    def __init__(self, master=None):
        QtWidgets.QMainWindow.__init__(self, master)
        self.setWindowTitle("Main Media Player")

        # Create a basic vlc instance
        self.instance = vlc.Instance()

        self.media = None

        # Create an empty vlc media player
        self.mediaplayer = self.instance.media_player_new()

        self.create_ui()
        self.data_queue = queue.Queue()
        self.is_paused = False

        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(100)
        self.timer.timeout.connect(self.update_ui)
        self.timer.timeout.connect(self.update_time_label)

    def create_ui(self):
        """Set up the user interface, signals & slots
        """
        self.widget = QtWidgets.QWidget(self)
        self.setCentralWidget(self.widget)

        # In this widget, the video will be drawn
        if platform.system() == "Darwin":  # for MacOS
            self.videoframe = QtWidgets.QMacCocoaViewContainer(0)
        else:
            self.videoframe = QtWidgets.QFrame()

        self.overlay = Overlay(self.videoframe)

        self.palette = self.videoframe.palette()
        self.palette.setColor(QtGui.QPalette.Window, QtGui.QColor(0, 0, 0))
        self.videoframe.setPalette(self.palette)
        self.videoframe.setAutoFillBackground(True)

        # Create the time display
        self.timelabel = QtWidgets.QLabel("00:00:00", self)

        # Create the position slider (QSlider)
        self.positionslider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        self.positionslider.setToolTip("Position")
        self.positionslider.setMaximum(1000)
        self.positionslider.sliderMoved.connect(self.set_position)
        # self.positionslider.sliderPressed.connect(self.set_position)
        self.positionslider.sliderMoved.connect(self.update_time_label)

        # Create the "previous frame" button
        self.previousframe = QtWidgets.QPushButton()
        self.previousframe.setFixedWidth(25)
        self.previousframe.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaSkipBackward))
        self.previousframe.clicked.connect(self.on_previous_frame)

        # Create the play button and connect it to the play/pause function
        self.playbutton = QtWidgets.QPushButton()
        self.playbutton.setFixedWidth(40)
        self.playbutton.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPlay))
        self.playbutton.clicked.connect(self.play_pause)

        # Create the "next frame" button
        self.nextframe = QtWidgets.QPushButton()
        self.nextframe.setFixedWidth(25)
        self.nextframe.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaSkipForward))
        self.nextframe.clicked.connect(self.on_next_frame)

        # Create the "Playback rate" label
        self.pb_rate_label = QtWidgets.QLabel("Playback rate: {}x".format(self.mediaplayer.get_rate()), self)

        # Create the "decrease playback rate" button
        self.decr_pb_rate = QtWidgets.QPushButton()
        self.decr_pb_rate.setFixedWidth(30)
        self.decr_pb_rate.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaSeekBackward))
        self.decr_pb_rate.clicked.connect(self.decr_mov_play_rate)

        # Create the stop button and connect it to the stop function
        self.stopbutton = QtWidgets.QPushButton()
        self.stopbutton.setFixedWidth(30)
        self.stopbutton.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaStop))
        self.stopbutton.clicked.connect(self.stop)

        # Create the "increase playback rate" button
        self.incr_pb_rate = QtWidgets.QPushButton()
        self.incr_pb_rate.setFixedWidth(30)
        self.incr_pb_rate.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaSeekForward))
        self.incr_pb_rate.clicked.connect(self.incr_mov_play_rate)

        self.top_control_box = QtWidgets.QHBoxLayout()

        # Add the time and position slider to the 1st controls layout
        self.top_control_box.addWidget(self.timelabel)
        self.top_control_box.addWidget(self.positionslider)

        self.bottom_control_box = QtWidgets.QHBoxLayout()

        # Add the buttons to the 2nd controls layout
        self.bottom_control_box.addWidget(self.previousframe)
        self.bottom_control_box.addWidget(self.playbutton)
        self.bottom_control_box.addWidget(self.nextframe)
        self.bottom_control_box.addWidget(self.pb_rate_label)
        self.bottom_control_box.addWidget(self.decr_pb_rate)
        self.bottom_control_box.addWidget(self.stopbutton)
        self.bottom_control_box.addWidget(self.incr_pb_rate)

        self.vboxlayout = QtWidgets.QVBoxLayout()

        self.vboxlayout.addWidget(self.videoframe)
        self.vboxlayout.addLayout(self.top_control_box)
        self.vboxlayout.addLayout(self.bottom_control_box)

        self.widget.setLayout(self.vboxlayout)

        menu_bar = self.menuBar()

        # File menu
        file_menu = menu_bar.addMenu("File")

        # Create submenu to start new processes from file menu
        new_menu = QtWidgets.QMenu("Launch", self)
        file_menu.addMenu(new_menu)
        new_plot_action = QtWidgets.QAction("New Plot Window", self)
        new_video_action = QtWidgets.QAction("New Video", self)
        new_menu.addAction(new_plot_action)
        new_menu.addAction(new_video_action)
        new_plot_action.triggered.connect(on_new_plot)
        new_video_action.triggered.connect(on_new_video)

        # Create actions to load a new media file and to close the app
        open_action = QtWidgets.QAction("Load Video", self)
        close_action = QtWidgets.QAction("Close App", self)
        file_menu.addAction(open_action)
        file_menu.addAction(close_action)
        open_action.triggered.connect(self.open_file)
        close_action.triggered.connect(sys.exit)

    def play_pause(self):
        """Toggle play/pause status
        """
        if self.mediaplayer.is_playing():
            signal = 'p'
            self.mediaplayer.pause()
            self.playbutton.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPlay))
            self.is_paused = True
            self.timer.stop()
        else:
            if self.mediaplayer.play() == -1:
                self.open_file()
                return

            signal = 'P'
            self.mediaplayer.play()
            self.playbutton.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPause))
            self.timer.start()
            self.is_paused = False

        # Reset the queue & send the appropriate signal, i.e., play/pause
        self.data_queue.queue.clear()
        self.data_queue.put('d')
        self.data_queue.put(signal)

    def stop(self):
        """Stop player
        """
        self.mediaplayer.stop()
        self.playbutton.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPlay))

        # Reset the time label back to 00:00:00
        time = QtCore.QTime(0, 0, 0, 0)
        self.timelabel.setText(time.toString())

        # Reset the queue
        self.data_queue.queue.clear()
        self.data_queue.put('d')
        self.data_queue.put('S')

        # Reset the media position slider
        self.positionslider.setValue(0)

        self.timer.stop()

    def on_next_frame(self):
        """Go forward one frame.

            The Python VLC binding next_frame function causes a:

            "direct3d11 vout display error: SetThumbNailClip failed"

            error when next_frame is called while the video is playing,
            so we are using our own fucntion to get the next frame.
        """
        # self.mediaplayer.next_frame()
        next_frame_time = self.mediaplayer.get_time() + self.mspf()

        # Reset the queue & put the next frame's time into the queue
        self.data_queue.queue.clear()
        self.data_queue.put('d')
        self.data_queue.put(next_frame_time)
        self.update_time_label()
        self.mediaplayer.set_time(next_frame_time)

    def on_previous_frame(self):
        """Go backward one frame"""
        next_frame_time = self.mediaplayer.get_time() - self.mspf()

        # Reset the queue & put the next frame's time into the queue
        self.data_queue.queue.clear()
        self.data_queue.put('d')
        self.data_queue.put(next_frame_time)
        self.update_time_label()
        self.mediaplayer.set_time(next_frame_time)

    def mspf(self):
        """Milliseconds per frame"""
        return int(1000 // (self.mediaplayer.get_fps() or 25))

    def incr_mov_play_rate(self):
        """Increase the movie play rate by a factor of 2."""
        if self.mediaplayer.get_rate() >= 64:
            return

        rate = self.mediaplayer.get_rate() * 2
        result = self.mediaplayer.set_rate(rate)
        if result == 0:
            self.data_queue.queue.clear()
            self.data_queue.put('d')
            self.data_queue.put('>')
            self.update_pb_rate_label()

    def decr_mov_play_rate(self):
        """Decrease the movie play rate by a factor of 2."""
        if self.mediaplayer.get_rate() <= 0.125:
            return

        rate = self.mediaplayer.get_rate() * 0.5
        result = self.mediaplayer.set_rate(rate)
        if result == 0:
            self.data_queue.queue.clear()
            self.data_queue.put('d')
            self.data_queue.put('<')
            self.update_pb_rate_label()

    def open_file(self):
        """Open a media file in a MediaPlayer
        """
        dialog_txt = "Choose Media File"
        filename = QtWidgets.QFileDialog.getOpenFileName(self, dialog_txt, os.path.expanduser('~'))
        if not filename[0]:
            return

        # getOpenFileName returns a tuple, so use only the actual file name
        self.media = self.instance.media_new(filename[0])

        # Put the media in the media player
        self.mediaplayer.set_media(self.media)

        # Parse the metadata of the file
        self.media.parse()

        # Set the title of the track as window title
        self.setWindowTitle("Main Media Player: {}".format(self.media.get_meta(0)))

        # The media player has to be 'connected' to the QFrame (otherwise the
        # video would be displayed in it's own window). This is platform
        # specific, so we must give the ID of the QFrame (or similar object) to
        # vlc. Different platforms have different functions for this.
        if platform.system() == "Linux":  # for Linux using the X Server
            self.mediaplayer.set_xwindow(int(self.videoframe.winId()))
        elif platform.system() == "Windows":  # for Windows
            self.mediaplayer.set_hwnd(int(self.videoframe.winId()))
        elif platform.system() == "Darwin":  # for MacOS
            self.mediaplayer.set_nsobject(int(self.videoframe.winId()))

        self.play_pause()

    def set_position(self):
        """Set the movie position according to the position slider.

        The vlc MediaPlayer needs a float value between 0 and 1, Qt uses
        integer variables, so you need a factor; the higher the factor,
        the more precise are the results (1000 should suffice).
        """
        # Set the media position to where the slider was dragged
        self.timer.stop()
        pos = self.positionslider.value()

        if pos >= 0:
            self.data_queue.queue.clear()
            self.data_queue.put('d')
            current_time = self.mediaplayer.get_time()

            # If the player is stopped, do not attempt to send a -1!!!
            if current_time == -1:
                self.timer.start()
                return
            self.data_queue.put(current_time)

        self.mediaplayer.set_position(pos * .001)
        self.timer.start()

    def update_ui(self):
        """Updates the user interface"""

        # Set the slider's position to its corresponding media position
        # Note that the setValue function only takes values of type int,
        # so we must first convert the corresponding media position.
        media_pos = int(self.mediaplayer.get_position() * 1000)
        self.positionslider.setValue(media_pos)

        if media_pos >= 0 and self.mediaplayer.is_playing():
            current_time = self.mediaplayer.get_time()
            self.data_queue.put(current_time)
        else:
            self.data_queue.queue.clear()

        # No need to call this function if nothing is played
        if not self.mediaplayer.is_playing():
            self.timer.stop()

            # After the video finished, the play button stills shows "Pause",
            # which is not the desired behavior of a media player.
            # This fixes that "bug".
            if not self.is_paused:
                self.stop()

    def update_time_label(self):
        mtime = QtCore.QTime(0, 0, 0, 0)
        self.time = mtime.addMSecs(self.mediaplayer.get_time())
        self.timelabel.setText(self.time.toString())

    def update_pb_rate_label(self):
        print(self.mediaplayer.get_rate())
        self.pb_rate_label.setText("Playback rate: {}x".format(str(self.mediaplayer.get_rate())))


def on_new_plot():
    """Launches a new PyQtGraph plot(s) window
    """
    subprocess.Popen(["python", "client.py"])


def on_new_video():
    """Launches a new PyQt5-based "mini" media player
    """
    if platform.system() == "Windows":
        subprocess.Popen(["python", "mini_player.py"], shell=True)
    else:
        subprocess.Popen(["python", "mini_player.py"])


def main():
    """Entry point for our simple vlc player
    """
    app = QtWidgets.QApplication(sys.argv)
    player = Player()

    _ = Server("localhost", 10000, player.data_queue)

    player.show()
    player.resize(640, 480)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Fri Jan 11 11:32:12 2019

@Author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
"""

import random

from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QBrush, QPainter, QPalette


class Overlay(QtWidgets.QWidget):

    def __init__(self, parent=None):
        QtWidgets.QWidget.__init__(self, parent)

        palette = QPalette(self.palette())
        palette.setColor(palette.Background, Qt.transparent)
        self.setPalette(palette)

        # Set (x, y) as its top-left corner and the given width and height
        parent_geometry = self.parent().geometry()
        top_left_x = parent_geometry.topLeft().x()
        top_left_y = parent_geometry.topLeft().y()
        width = parent_geometry.width()
        height = parent_geometry.height()
        self.setGeometry(top_left_x, top_left_y, width, height)

    def paintEvent(self, event):
        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.fillRect(event.rect(), QBrush(QColor(255, 255, 255, 10)))
        painter.setPen(Qt.red)

        size = self.parent().size()
        # print(size)
        for _ in range(5000):
            x = random.randint(1, size.width()-1)
            y = random.randint(1, size.height()-1)
            painter.drawPoint(x, y)

        painter.end()

    def showEvent(self, event):
        self.timer = self.startTimer(50)
        self.counter = 0

    def timerEvent(self, event):
        self.counter += 1
        self.update()



"""
This module contains all the dialogs to present to a user.

@author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
"""

import sys

from PyQt5 import QtWidgets, QtCore, QtGui

class DataDialog(QtWidgets.QDialog):
    """A dialog for presenting and selecting columns from a csv"""

    def __init__(self, column_names, parent=None, checkbox=False):
        super(DataDialog, self).__init__(parent)

        self.layout = QtWidgets.QVBoxLayout()
        self.listWidget = QtWidgets.QListWidget()
        self.listWidget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.listWidget.setGeometry(QtCore.QRect(10, 10, 211, 291))

        # Fill QList with column names that were passed in
        for column_name in column_names:
            item = QtWidgets.QListWidgetItem(column_name)
            self.listWidget.addItem(item)

        self.button_box = QtWidgets.QHBoxLayout()
        self.ok_button = QtWidgets.QPushButton("Ok")
        self.ok_button.clicked.connect(self.handle_ok_button)
        self.cancel_button = QtWidgets.QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.handle_cancel_button)

        self.button_box.addWidget(self.ok_button)
        self.button_box.addWidget(self.cancel_button)

        self.listWidget.itemSelectionChanged.connect(self.update_selected_items)

        self.layout.addWidget(self.listWidget)
        self.layout.addLayout(self.button_box)

        # Add a checkbox to the dialog for the user to toggle
        # if they want to customize behavior colors
        if checkbox:
            self.choose_colors_cb = QtWidgets.QCheckBox("Choose behavior colors", self)
            self.layout.addWidget(self.choose_colors_cb)
            self.choose_colors_cb.stateChanged.connect(self.handle_checkbox)

            self.choose_colors = False

        self.setLayout(self.layout)

        # Type of exit, (i.e. close, ok, cancel)
        # close: 0; cancel: 0; ok: 1
        self.exit_status = 0

        self.selected_items = []

    def closeEvent(self, event):
        """Handles what happens with items when dialog is closed"""

        if self.exit_status != 1:
            self.selected_items = []

        print(self.selected_items)
        print("Data Dialog closed!")

    def update_selected_items(self):
        """Keeps the selected items list updated"""

        items = self.listWidget.selectedItems()
        self.selected_items = [item.text() for item in items]

    def handle_ok_button(self):
        """Handles what happens when Ok button is clicked"""

        self.exit_status = 1
        self.close()

    def handle_cancel_button(self):
        """Handles what happens when Cancel button is clicked"""

        self.exit_status = 0
        print("Canceled!")
        self.close()

    def handle_checkbox(self):
        """Handles what happens when checkbox is toggled"""

        self.choose_colors = not self.choose_colors

class ColorsDialog(QtWidgets.QDialog):
    """A dialog for selecting behavior colors"""

    def __init__(self, behaviors, parent=None):
        super(ColorsDialog, self).__init__(parent)

        self.layout = QtWidgets.QVBoxLayout()

        for behavior in behaviors:
            behavior_button = QtWidgets.QPushButton(behavior)
            behavior_button.clicked.connect(self.handle_behavior_button)
            # behavior_button.setDefault(False)
            behavior_button.setAutoDefault(False)

            self.layout.addWidget(behavior_button)

        self.ok_button = QtWidgets.QPushButton("Ok")
        self.ok_button.clicked.connect(self.on_ok)

        self.cancel_button = QtWidgets.QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.on_cancel)

        self.button_box = QtWidgets.QHBoxLayout()
        self.button_box.addWidget(self.ok_button)
        self.button_box.addStretch()
        self.button_box.addWidget(self.cancel_button)

        self.layout.addLayout(self.button_box)
        self.setLayout(self.layout)

        # Type of exit, (i.e. close, ok, cancel)
        # close: 0; cancel: 0; ok: 1
        self.exit_status = 0

        self.behavior_colors = {}

    def closeEvent(self, event):
        """Handles what happens with items when dialog is closed"""

        if self.exit_status != 1:
            self.behavior_colors = {}

        print("Colors Dialog closed!")

    def handle_behavior_button(self):
        """Displays a dialog for selecting colour values"""

        # Get the name of the button (i.e. the behavior) that was clicked
        button_name = self.sender().text()

        # Get the name of the button (i.e. the behavior) that was clicked
        button_name = self.sender().text()

        color_dialog = QtWidgets.QColorDialog(self)

        # Let users choose a color with a level of transparency
        color_dialog.setOption(QtWidgets.QColorDialog.ShowAlphaChannel)
        options = color_dialog.options()
        color = color_dialog.getColor(options=options)

        self.behavior_colors[button_name] = color.getRgb()

    def on_ok(self):
        """Handles what happens when Ok button is clicked"""

        self.exit_status = 1
        self.close()

    def on_cancel(self):
        """Handles what happens when Cancel button is clicked"""

        self.exit_status = 0
        print("Canceled!")
        self.close()

class AxisDialog(QtWidgets.QDialog):

    def __init__(self, parent=None):
        super(AxisDialog, self).__init__(parent)

        self.vert_layout = QtWidgets.QVBoxLayout()

        self.hbox = QtWidgets.QHBoxLayout()

        lower_axis_label = QtWidgets.QLabel("Lower bound:")
        self.lower_axis = QtWidgets.QLineEdit()
        self.lower_axis.setValidator(QtGui.QIntValidator())

        upper_axis_label = QtWidgets.QLabel("Upper bound:")
        self.upper_axis = QtWidgets.QLineEdit()
        self.upper_axis.setValidator(QtGui.QIntValidator())

        self.hbox.addWidget(lower_axis_label)
        self.hbox.addWidget(self.lower_axis)
        self.hbox.addWidget(upper_axis_label)
        self.hbox.addWidget(self.upper_axis)

        self.vert_layout.addLayout(self.hbox)

        self.ok_button = QtWidgets.QPushButton("Ok")
        self.ok_button.clicked.connect(self.on_ok)

        self.cancel_button = QtWidgets.QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.on_cancel)

        self.button_box = QtWidgets.QHBoxLayout()
        self.button_box.addWidget(self.ok_button)
        self.button_box.addWidget(self.cancel_button)

        self.vert_layout.addLayout(self.button_box)
        self.setLayout(self.vert_layout)

        self.exit_status = 0

        self.lower_bound = None
        self.upper_bound = None

    def closeEvent(self, event):
        """Handles what happens with items when dialog is closed"""

        if self.exit_status != 1:
            self.lower_bound = None
            self.upper_bound = None
        else:
            self.lower_bound = self.lower_axis.text()
            self.upper_bound = self.upper_axis.text()

        print("lower: {}, upper: {}".format(self.lower_bound, self.upper_bound))

    def on_ok(self):
        """Handles what happens when Ok button is clicked"""

        self.exit_status = 1
        self.close()

    def on_cancel(self):
        """Handles what happens when Cancel button is clicked"""

        self.exit_status = 0
        print("Canceled!")
        self.close()

def main():
    """This is just for testing that the DataDialog class works"""

    app = QtWidgets.QApplication(sys.argv)
    # data_dialog = DataDialog([str(i) for i in range(4)], checkbox=True)
    # data_dialog.show()

    # axis_dialog = AxisDialog()
    # axis_dialog.show()

    color_dialog = ColorsDialog([str(i) for i in range(4)])
    color_dialog.show()
    app.exec_()
    # print(data_dialog.selected_items)

if __name__ == "__main__":
    main()



"""This module contains a bare-bones vlc player class to stream videos.
"""

import os
import sys
import queue
import platform

from PyQt5 import QtWidgets, QtGui, QtCore
import vlc
from client import Client


class MiniPlayer(QtWidgets.QMainWindow):
    """Stripped-down PyQt5-based media player class to sync with another video.
    """

    def __init__(self, data_queue, master=None):
        QtWidgets.QMainWindow.__init__(self, master)
        self.setWindowTitle("Mini Player")
        self.statusbar = self.statusBar()
        self.statusbar.showMessage("Ready")

        # Create a basic vlc instance
        self.instance = vlc.Instance()

        self.media = None

        # Create an empty vlc media player
        self.mediaplayer = self.instance.media_player_new()

        self.init_ui()
        self.open_file()

        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(10)
        self.timer.timeout.connect(self.update_ui)

        self.data_queue = data_queue
        self.timer.start()

    def init_ui(self):
        """Set up the user interface
        """
        self.widget = QtWidgets.QWidget(self)
        self.setCentralWidget(self.widget)

        # In this widget, the video will be drawn
        if platform.system() == "Darwin":  # for MacOS
            self.videoframe = QtWidgets.QMacCocoaViewContainer(0)
        else:
            self.videoframe = QtWidgets.QFrame()

        self.palette = self.videoframe.palette()
        self.palette.setColor(QtGui.QPalette.Window, QtGui.QColor(0, 0, 0))
        self.videoframe.setPalette(self.palette)
        self.videoframe.setAutoFillBackground(True)

        self.vboxlayout = QtWidgets.QVBoxLayout()
        self.vboxlayout.addWidget(self.videoframe)
        self.widget.setLayout(self.vboxlayout)

    def open_file(self):
        """Open a media file in a MediaPlayer
        """
        dialog_txt = "Choose Media File"
        filename = QtWidgets.QFileDialog.getOpenFileName(self, dialog_txt, os.path.expanduser('~'))
        if not filename[0]:
            return

        # getOpenFileName returns a tuple, so use only the actual file name
        self.media = self.instance.media_new(filename[0])

        # Put the media in the media player
        self.mediaplayer.set_media(self.media)

        # Parse the metadata of the file
        self.media.parse()

        # Set the title of the track as the window title
        self.setWindowTitle("{}".format(self.media.get_meta(0)))

        # The media player has to be 'connected' to the QFrame (otherwise the
        # video would be displayed in it's own window). This is platform
        # specific, so we must give the ID of the QFrame (or similar object) to
        # vlc. Different platforms have different functions for this
        if platform.system() == "Linux":  # for Linux using the X Server
            self.mediaplayer.set_xwindow(int(self.videoframe.winId()))
        elif platform.system() == "Windows":  # for Windows
            self.mediaplayer.set_hwnd(int(self.videoframe.winId()))
        elif platform.system() == "Darwin":  # for MacOS
            self.mediaplayer.set_nsobject(int(self.videoframe.winId()))

        # Start playing the video as soon as it loads
        self.mediaplayer.play()

    def update_ui(self):
        self.update_statusbar()

        try:
            val = self.data_queue.get(block=False)
        except queue.Empty:
            return

        if val == '<':
            self.mediaplayer.set_rate(self.mediaplayer.get_rate() * 0.5)
            return
        if val == '>':
            self.mediaplayer.set_rate(self.mediaplayer.get_rate() * 2)
            return
        if val == 'P':
            self.mediaplayer.play()
            return
        elif val == 'p':
            self.mediaplayer.pause()
            return
        elif val == 'S':
            self.mediaplayer.stop()
            return
        else:
            val = int(val)
            if val != self.mediaplayer.get_time():
                self.mediaplayer.set_time(val)

    def update_statusbar(self):
        mtime = QtCore.QTime(0, 0, 0, 0)
        time = mtime.addMSecs(self.mediaplayer.get_time())
        self.statusbar.showMessage(time.toString())


def main():
    """Entry point for our simple vlc player
    """
    app = QtWidgets.QApplication(sys.argv)

    data_queue = queue.Queue()

    player = MiniPlayer(data_queue)
    player.show()
    player.resize(480, 480)

    _ = Client("localhost", 10000, data_queue)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()
